# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the NoopWorkflow class."""

from debusine.server.workflows import NoopWorkflow
from debusine.server.workflows.models import BaseWorkflowData
from debusine.server.workflows.tests.helpers import WorkflowTestCase


class NoopWorkflowTests(WorkflowTestCase[NoopWorkflow]):
    """Unit tests for NoopWorkflow class."""

    def test_create(self) -> None:
        self.assertIsInstance(self.task.data, BaseWorkflowData)

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Django pages URL's for Debusine."""

from django.urls import URLPattern, URLResolver, include, path
from django.views.generic import RedirectView

service_urlpatterns: list[URLPattern | URLResolver] = [
    path(
        "groups/",
        include("debusine.web.urls.groups", namespace="groups"),
    ),
]

# Scoped URLs
urlpatterns = [
    path("-/", include(service_urlpatterns)),
    # TODO: Preserved for compatibility redirects only
    path(
        "accounts/oidc_callback/<name>/",
        RedirectView.as_view(
            pattern_name="signon:oidc_callback", permanent=True
        ),
    ),
    path(
        "accounts/bind_identity/<name>/",
        RedirectView.as_view(
            pattern_name="signon:bind_identity", permanent=True
        ),
    ),
    path(
        "",
        include("debusine.web.urls.scopes", namespace="scopes"),
    ),
    path(
        "",
        include("debusine.web.urls.workspaces", namespace="workspaces"),
    ),
    path(
        "",
        include(
            "debusine.web.urls.workflow_templates",
            namespace="workflow-templates",
        ),
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the forms."""

import datetime as dt
from collections.abc import Iterable
from typing import Any, ClassVar, override
from unittest import mock

from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.forms.fields import ChoiceField, IntegerField, InvalidJSONInput
from django.test import RequestFactory

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
    TaskTypes,
)
from debusine.db.models import (
    Collection,
    CollectionRelation,
    Group,
    Token,
    User,
    Workspace,
)
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.forms import (
    BaseFormSetBase,
    BootstrapMixin,
    CollectionCreateForm,
    CollectionFormBase,
    CollectionRelationCreateForm,
    CollectionUpdateForm,
    DaysField,
    GroupAddUserForm,
    TaskConfigurationInspectorForm,
    TokenForm,
    WorkflowFilterForm,
    WorkspaceChoiceField,
    WorkspaceForm,
    WorkspaceInheritanceForm,
    WorkspaceInheritanceFormSet,
    YAMLField,
)
from debusine.web.views.ui.workspaces import WorkspaceUI


class BootstrapMixinForm(BootstrapMixin, forms.Form):
    """Class to test BootstrapMixinTestCase."""

    existing_class = "class1"
    char_field = forms.CharField()
    char_field_required = forms.CharField(required=True)
    char_field_required_with_suffix = forms.CharField(
        required=True, label_suffix="something"
    )
    char_field_extra_class = forms.CharField(
        widget=forms.TextInput(attrs={"class": existing_class})
    )
    choice_field = forms.ChoiceField(choices=[("1", "One"), ("2", "Two")])
    choice_field_extra_class = forms.ChoiceField(
        choices=[], widget=forms.Select(attrs={"class": existing_class})
    )
    datetime_field = forms.DateTimeField(
        widget=forms.DateTimeInput(attrs={"class": existing_class})
    )
    file_field = forms.FileField()

    date_field = forms.DateField(
        widget=forms.DateInput(attrs={"class": existing_class})
    )
    boolean_field = forms.BooleanField(
        widget=forms.CheckboxInput(attrs={"class": existing_class})
    )
    days_field = DaysField()
    integer_field = IntegerField()


class BootstrapMixinTests(TestCase):
    """Tests for BootstrapMixin."""

    form: ClassVar[BootstrapMixinForm]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.form = BootstrapMixinForm()

    def test_charfield_bootstrap_class(self) -> None:
        """CharField has the correct Bootstrap class."""  # noqa: D403
        char_widget_class = self.form.fields["char_field"].widget.attrs.get(
            "class"
        )
        self.assertEqual(char_widget_class, "form-control")

    def test_charfield_bootstrap_class_add(self) -> None:
        """CharField has the correct classes."""  # noqa: D403
        char_widget_class = self.form.fields[
            "char_field_extra_class"
        ].widget.attrs.get("class")
        self.assertEqual(
            char_widget_class,
            f"{BootstrapMixinForm.existing_class} form-control",
        )

    def test_choicefield_bootstrap_class(self) -> None:
        """ChoiceField has the correct Bootstrap class."""  # noqa: D403
        choice_widget_class = self.form.fields["choice_field"].widget.attrs.get(
            "class"
        )
        self.assertEqual(choice_widget_class, "form-select")

    def test_choicefield_bootstrap_class_add(self) -> None:
        """ChoiceField has the correct classes."""  # noqa: D403
        choice_widget_class = self.form.fields[
            "choice_field_extra_class"
        ].widget.attrs.get("class")
        self.assertEqual(
            choice_widget_class,
            f"{BootstrapMixinForm.existing_class} form-select",
        )

    def test_filefield_bootstrap_class(self) -> None:
        """FileField has the correct Bootstrap class."""  # noqa: D403
        file_widget_class = self.form.fields["file_field"].widget.attrs.get(
            "class"
        )
        self.assertEqual(file_widget_class, "form-control")

    def test_datetimefield_bootstrap_class(self) -> None:
        """Test DateField retains its original class."""  # noqa: D403
        date_widget_class = self.form.fields["datetime_field"].widget.attrs.get(
            "class"
        )
        self.assertEqual(
            date_widget_class,
            f"{BootstrapMixinForm.existing_class} form-control",
        )

    def test_datefield_retain_existing_class(self) -> None:
        """
        Test DateField retains its original class.

        DateField is not modified: nothing is added.
        """  # noqa: D403
        date_widget_class = self.form.fields["date_field"].widget.attrs.get(
            "class"
        )
        self.assertEqual(
            date_widget_class, f"{BootstrapMixinForm.existing_class}"
        )

    def test_daysfield_bootstrap_class(self) -> None:
        """Test DaysField has form-control."""
        days_widget_class = self.form.fields["days_field"].widget.attrs.get(
            "class"
        )
        self.assertEqual(days_widget_class, "form-control")

    def test_integerfield_bootstrap_class(self) -> None:
        integer_widget_class = self.form.fields[
            "integer_field"
        ].widget.attrs.get("class")
        self.assertEqual(integer_widget_class, "form-control")

    def test_booleanfield_bootstrap_class(self) -> None:
        """Test BooleanField has form-check-input."""
        boolean_widget_class = self.form.fields[
            "boolean_field"
        ].widget.attrs.get("class")
        self.assertEqual(
            boolean_widget_class,
            f"{BootstrapMixinForm.existing_class} form-check-input",
        )

    def test_asterisk_set_if_required(self) -> None:
        """Test asterisk is set if field is required."""
        char_field_required = self.form.fields["char_field_required"]
        self.assertEqual(char_field_required.label_suffix, " *")

    def test_asterisk_added_if_required(self) -> None:
        """Test asterisk is added if field is required."""
        char_field_required = self.form.fields[
            "char_field_required_with_suffix"
        ]
        self.assertEqual(char_field_required.label_suffix, "something *")


class YAMLFieldTests(TestCase):
    """Tests for :py:class:`YAMLField`."""

    def test_init_with_mutable(self) -> None:
        field = YAMLField(empty_value=3)
        self.assertEqual(field.empty_value, 3)
        field = YAMLField(empty_value=dict)
        self.assertEqual(field.empty_value, {})
        field = YAMLField(empty_value=list)
        self.assertEqual(field.empty_value, [])

    def test_to_python(self) -> None:
        field = YAMLField(empty_value=dict)
        for value, expected in (
            (None, {}),
            ("", {}),
            ("3", 3),
            ("test", "test"),
            ("test: 1", {"test": 1}),
        ):
            with self.subTest(value=value):
                self.assertEqual(field.to_python(value), expected)

    def test_to_python_disabled(self) -> None:
        field = YAMLField(empty_value=dict, disabled=True)
        for value in (None, "", "3", "test", "test: 1"):
            with self.subTest(value=value):
                self.assertEqual(field.to_python(value), value)

    def test_bound_data(self) -> None:
        field = YAMLField(empty_value=dict)
        for value, expected in (
            (None, None),
            ("", None),
            ("3", 3),
            ("test", "test"),
            ("test: 1", {"test": 1}),
            ("{", "{"),
        ):
            with self.subTest(value=value):
                self.assertEqual(field.bound_data(value, 42), expected)

    def test_bound_data_disabled(self) -> None:
        field = YAMLField(empty_value=dict, disabled=True)
        for value in (None, "", "3", "test", "test: 1"):
            with self.subTest(value=value):
                self.assertEqual(field.bound_data(value, 42), 42)

    def test_prepare_value(self) -> None:
        field = YAMLField(empty_value=dict)
        for value, expected in (
            (None, "{}\n"),
            ("", "{}\n"),
            (3, "3\n...\n"),
            ("test", "test\n...\n"),
            ({"test": 1}, "test: 1\n"),
            ("{", "'{'\n"),
            (InvalidJSONInput("{"), "{"),
        ):
            with self.subTest(value=repr(value)):
                self.assertEqual(field.prepare_value(value), expected)

    def test_has_changed(self) -> None:
        field = YAMLField(empty_value=dict)
        for initial, data, expected in (
            (None, None, True),
            (None, "", True),
            ([1, 2, 3], [1, 2, 3], False),
            ([1, 2, 3], [1, 2, 4], True),
            ({"one": 1, "two": 2}, {"two": 2, "one": 1}, False),
            ({"one": 1, "two": 2}, {"two": 3, "one": 1}, True),
        ):
            with self.subTest(initial=initial, data=data):
                self.assertEqual(field.has_changed(initial, data), expected)

    def test_invalid_yaml(self) -> None:
        field = YAMLField()
        invalid_yaml = "Task: : value"

        with self.assertRaises(ValidationError):
            field.to_python(invalid_yaml)


class TokenFormTests(TestCase):
    """Tests for TokenForm."""

    user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = get_user_model().objects.create_user(
            username="testuser", password="testpass"
        )

    def test_form_initialization(self) -> None:
        """Form initialization set form.user."""
        form = TokenForm(user=self.user)
        self.assertEqual(form.user, self.user)
        self.assertTrue(form.fields["enabled"].initial)

    def test_form_save(self) -> None:
        """Form save the Token."""
        form = TokenForm({"comment": "Test Comment"}, user=self.user)
        self.assertTrue(form.is_valid())

        token = form.save()

        self.assertEqual(token.user, self.user)
        self.assertEqual(token.comment, "Test Comment")

    def test_form_save_commit_false(self) -> None:
        """Form does not save the token: commit=False."""
        form = TokenForm({"comment": "Test Comment"}, user=self.user)
        self.assertTrue(form.is_valid())

        token = form.save(commit=False)

        self.assertEqual(token.user, self.user)
        self.assertEqual(token.comment, "Test Comment")
        # Ensure token is not saved to the database yet
        with self.assertRaises(Token.DoesNotExist):
            Token.objects.get(comment="Test Comment")

    def test_form_validation(self) -> None:
        """Form return an error if the "comment" is not valid: too long."""
        assert Token.comment.field.max_length is not None
        form = TokenForm(
            {"comment": "x" * (Token.comment.field.max_length + 1)},
            user=self.user,
        )
        self.assertFalse(form.is_valid())
        self.assertIn("comment", form.errors)

    def test_comment_allowed_empty(self) -> None:
        """Form validates if comment is empty."""
        form = TokenForm({"comment": ""}, user=self.user)
        self.assertTrue(form.is_valid())


class WorkspaceChoiceFieldTests(TestCase):
    """Tests for WorkspaceChoiceField."""

    user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = get_user_model().objects.create_user(
            username="testuser", password="testpass"
        )

    def test_label(self) -> None:
        """The label is the workspace name."""
        workspace_choice_field = WorkspaceChoiceField(user=self.user)

        labels = [
            label for value, label in workspace_choice_field.widget.choices
        ]

        self.assertEqual(
            labels,
            [
                workspace_choice_field.empty_label,
                Workspace.objects.earliest("id").name,
            ],
        )

    def test_order_by(self) -> None:
        """The field's queryset is ordered by name."""
        workspace_choice_field = WorkspaceChoiceField(user=None)

        assert workspace_choice_field.queryset is not None
        self.assertEqual(
            workspace_choice_field.queryset.query.order_by, ("name",)
        )

    def test_choices_no_user(self) -> None:
        """Choices for if no user: only the public workspaces."""
        default_workspace = Workspace.objects.earliest("id")
        default_workspace.public = False
        default_workspace.save()

        workspace = self.playground.create_workspace(
            name="It is public", public=True
        )

        workspace_choice_field = WorkspaceChoiceField(user=None)

        # Blank choice and public workspace
        self.assertEqual(
            list(workspace_choice_field.choices),  # type: ignore[arg-type]
            [
                ("", workspace_choice_field.empty_label),
                (workspace.id, workspace.name),
            ],
        )

    def test_choices_with_user(self) -> None:
        """Private workspace included in the choices: user has access."""
        workspace = Workspace.objects.earliest("id")
        workspace.public = False
        workspace.save()

        workspace_choice_field = WorkspaceChoiceField(user=self.user)

        self.assertEqual(
            list(workspace_choice_field.choices),  # type: ignore[arg-type]
            [
                ("", workspace_choice_field.empty_label),
                (workspace.pk, workspace.name),
            ],
        )


class WorkflowFilterFormTests(TestCase):
    """Tests for WorkflowFilterForm."""

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    def setUp(self) -> None:
        super().setUp()

        self.form = WorkflowFilterForm()

    def test_statuses_field(self) -> None:
        """Test statuses field."""
        self.assertEqual(
            self.form.fields["statuses"].widget.choices,
            [
                ('pending', 'Pending'),
                (
                    'Running',
                    [
                        (
                            'running__needs_input',
                            'Running (children: Needs Input)',
                        ),
                        ('running__running', 'Running (children: Running)'),
                        ('running__waiting', 'Running (children: Waiting)'),
                        ('running__pending', 'Running (children: Pending)'),
                        ('running__aborted', 'Running (children: Aborted)'),
                        ('running__completed', 'Running (children: Completed)'),
                        ('running__blocked', 'Running (children: Blocked)'),
                        ('running__any', 'Running (children: Any)'),
                    ],
                ),
                ('completed', 'Completed'),
                ('aborted', 'Aborted'),
                ('blocked', 'Blocked'),
            ],
        )

    def test_runtime_statuses_field(self) -> None:
        """Test runtime_statuses field."""
        self.assertEqual(
            self.form.fields["runtime_statuses"].widget.choices,
            [
                ("needs_input", "Needs Input"),
                ("running", "Running"),
                ("waiting", "Waiting"),
                ("pending", "Pending"),
                ("aborted", "Aborted"),
                ("completed", "Completed"),
                ("blocked", "Blocked"),
            ],
        )

    def test_results_field(self) -> None:
        """Test results field."""
        self.assertEqual(
            self.form.fields["results"].widget.choices,
            [
                ("error", "Error"),
                ("failure", "Failure"),
                ("skipped", "Skipped"),
                ("success", "Success"),
            ],
        )

    def test_with_failed_work_requests_failed_field(self) -> None:
        """Test with_failed_work_requests field."""
        self.assertIn("with_failed_work_requests", self.form.fields)


class DaysFieldTests(TestCase):
    """Tests for DaysField."""

    def test_prepare_value(self) -> None:
        """Test DaysField.prepare_value."""
        field = DaysField()
        for orig, prepared in [
            ("0", "0"),
            (1, 1),
            (dt.timedelta(days=4), 4),
            (None, None),
        ]:
            with self.subTest(orig=repr(orig)):
                self.assertEqual(field.prepare_value(orig), prepared)

    def test_to_python(self) -> None:
        """Test DaysField.to_python."""
        field = DaysField()
        for orig, py in [
            ("0", dt.timedelta(days=0)),
            (1, dt.timedelta(days=1)),
            ("", None),
            (None, None),
        ]:
            with self.subTest(orig=orig):
                self.assertEqual(field.to_python(orig), py)


class WorkspaceFormTests(TestCase):
    """Tests for WorkspaceForm."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_labels(self) -> None:
        form = WorkspaceForm(instance=self.scenario.workspace)
        self.assertEqual(
            form["expiration_delay"].help_text,
            "If unset, workspace is permanent. Otherwise, the workspace gets "
            "removed after the given period of inactivity (expressed as a "
            "number of days after the completion of the last work request)",
        )
        self.assertEqual(
            form["default_expiration_delay"].help_text,
            "Number of days that new artifacts and work requests are kept "
            "in the workspace before being expired (0 means no expiration)",
        )

    def test_values_from_instances(self) -> None:
        """Test form field values."""
        form = WorkspaceForm(instance=self.scenario.workspace)
        self.assertTrue(form["public"].initial)
        self.assertIsNone(form["expiration_delay"].initial)
        self.assertEqual(
            form["default_expiration_delay"].initial, dt.timedelta(days=0)
        )

    def test_values_from_data(self) -> None:
        """Test setting form field values."""
        ws = self.scenario.workspace
        form = WorkspaceForm(
            data={
                "public": "false",
                "expiration_delay": "7",
                "default_expiration_delay": "14",
            },
            instance=ws,
        )
        self.assertTrue(form.is_valid())
        form.save()

        ws.refresh_from_db()
        self.assertFalse(ws.public)
        self.assertEqual(ws.expiration_delay, dt.timedelta(days=7))
        self.assertEqual(ws.default_expiration_delay, dt.timedelta(days=14))


class WorkspaceInheritanceFormTests(TestCase):
    """Tests for :py:class:`WorkspaceInheritanceForm`."""

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    def setUp(self) -> None:
        """Set up a current request for tests."""
        super().setUp()
        factory = RequestFactory()
        self.request = factory.get("/")

    def get_helper(self, workspace: Workspace) -> WorkspaceUI:
        """Get the UI helper for a workspace."""
        return WorkspaceUI.get(self.request, workspace)

    def get_form(
        self, workspace: Workspace, **kwargs: Any
    ) -> WorkspaceInheritanceForm:
        """Instantiate a form from a workspace and form args."""
        return WorkspaceInheritanceForm(
            workspace_ui=self.get_helper(workspace), **kwargs
        )

    def test_choices(self) -> None:
        """``parent`` is a choice field with the expected workspaces."""
        ws1 = self.playground.create_workspace(name="ws1")
        ws2 = self.playground.create_workspace(name="ws2")
        self.get_helper(self.scenario.workspace).candidate_parents = [ws1, ws2]
        form = self.get_form(self.scenario.workspace)
        self.assertEqual(
            list(form.fields["parent"].widget.choices),
            [
                (None, "None"),
                (ws1.id, str(ws1)),
                (ws2.id, str(ws2)),
            ],
        )


class WorkspaceInheritanceFormSetTests(TestCase):
    """Tests for :py:class:`WorkspaceInheritanceFormSet`."""

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    def setUp(self) -> None:
        """Set up a current request for tests."""
        super().setUp()
        factory = RequestFactory()
        self.request = factory.get("/")

    def get_helper(self, workspace: Workspace) -> WorkspaceUI:
        """Get the UI helper for a workspace."""
        return WorkspaceUI.get(self.request, workspace)

    def get_formset_factory(
        self, workspace: Workspace, **kwargs: Any
    ) -> type[BaseFormSetBase[WorkspaceInheritanceForm]]:
        """Instantiate a form from a workspace and form args."""
        return WorkspaceInheritanceFormSet.formset_factory(
            workspace_ui=self.get_helper(workspace), **kwargs
        )

    def test_formset_factory(self) -> None:
        formset_class = self.get_formset_factory(self.scenario.workspace)
        formset = formset_class()
        self.assertEqual(formset.initial_form_count(), 0)
        self.assertEqual(formset.total_form_count(), 1)
        self.assertTrue(formset.can_order)
        self.assertTrue(formset.can_delete)
        self.assertIsInstance(formset.forms[0], WorkspaceInheritanceForm)
        self.assertEqual(
            formset.forms[0].workspace_ui,
            self.get_helper(self.scenario.workspace),
        )


class GroupAddUserFormTests(TestCase):
    """Tests for GroupAddUserForm."""

    scenario = scenarios.DefaultContext(set_current=True)

    group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("test")

    def _form(self, username: Any) -> GroupAddUserForm:
        """Instantiate the form."""
        return GroupAddUserForm(group=self.group, data={"username": username})

    def test_username_required(self) -> None:
        """A valid username passes validation."""
        form = self._form(None)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {"username": ["This field is required."]})

    def test_valid_username(self) -> None:
        """A valid username passes validation."""
        form = self._form(self.scenario.user.username)
        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data["username"], self.scenario.user.username
        )

    def test_invalid_username(self) -> None:
        """Error on invalid usernames."""
        form = self._form("does-not-exist")
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors, {"username": ["User does-not-exist does not exist"]}
        )

    def test_user_already_member(self) -> None:
        """Existing members are not accepted."""
        self.group.add_user(self.scenario.user)
        form = self._form(self.scenario.user.username)
        self.assertFalse(form.is_valid())
        self.assertEqual(list(form.errors.keys()), ["username"])
        self.assertEqual(
            form.errors,
            {"username": ["User is already a member of debusine/test"]},
        )


class TaskConfigurationInspectorFormTests(TestCase):
    """Tests for :py:class:`TaskConfigurationInspectorForm`."""

    scenario = scenarios.DefaultContext(set_current=True)
    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.scenario.workspace.collections.get(
            name="default", category=CollectionCategory.TASK_CONFIGURATION
        )

    @classmethod
    def add_config(cls, entry: DebusineTaskConfiguration) -> None:
        """Add a config entry to config_collection."""
        cls.collection.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=cls.scenario.user,
            data=entry,
        )

    def test_empty_collection(self) -> None:
        form = TaskConfigurationInspectorForm(collection=self.collection)
        unspecified = ('', '(unspecified)')
        self.assertEqual(form.fields["task"].widget.choices, [])
        self.assertEqual(form.fields["subject"].widget.choices, [unspecified])
        self.assertEqual(form.fields["context"].widget.choices, [unspecified])

    def test_choices(self) -> None:
        self.add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop"
            )
        )
        self.add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="sbuild"
            )
        )
        self.add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", context="context1"
            )
        )
        self.add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                context="context2",
                subject="subject1",
            )
        )
        self.add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="sbuild",
                context="context1",
                subject="subject2",
            )
        )

        form = TaskConfigurationInspectorForm(collection=self.collection)
        unspecified = ('', '(unspecified)')
        self.assertEqual(
            form.fields["task"].widget.choices,
            [
                ('Worker:noop', 'Worker:noop'),
                ('Worker:sbuild', 'Worker:sbuild'),
            ],
        )
        self.assertEqual(
            form.fields["subject"].widget.choices,
            [unspecified, ('subject1', 'subject1'), ('subject2', 'subject2')],
        )
        self.assertEqual(
            form.fields["context"].widget.choices,
            [unspecified, ('context1', 'context1'), ('context2', 'context2')],
        )


class CollectionFormBaseTests(TestCase):
    """Tests for :py:class:`CollectionFormBase`."""

    scenario = scenarios.DefaultContext(set_current=True)
    sid: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )

    def make_form_class(self, **meta: Any) -> type[CollectionFormBase]:
        """Create a concrete subclass of CollectionFormBase."""
        meta.setdefault(
            "fields",
            [
                "name",
                "category",
                "full_history_retention_period",
                "metadata_only_retention_period",
                "data",
            ],
        )
        meta.setdefault(
            "field_classes",
            {
                "full_history_retention_period": DaysField,
                "metadata_only_retention_period": DaysField,
            },
        )

        class ConcreteForm(CollectionFormBase):
            """Concrete version of CollectionFormBase used for testing."""

            Meta = type("Meta", (), {"model": Collection, **meta})

        return ConcreteForm

    def form(self, **meta: Any) -> CollectionFormBase:
        """Instantiate a test form."""
        form_class = self.make_form_class(**meta)
        form = form_class()
        # Simulate being inside full_clean
        form.cleaned_data = {}
        return form

    def test_check_timedelta_units(self) -> None:
        form = self.form()
        self.assertRegex(
            str(form.fields["full_history_retention_period"].help_text),
            r".+ \(in days\)$",
        )
        self.assertRegex(
            str(form.fields["metadata_only_retention_period"].help_text),
            r".+ \(in days\)$",
        )
        self.assertNotRegex(
            str(form.fields["name"].help_text), r".+ \(in days\)$"
        )

    def test_check_timedelta_units_without_fields(self) -> None:
        form = self.form(
            fields=["name", "category", "metadata_only_retention_period"]
        )
        self.assertNotIn("full_history_retention_period", form.fields)
        self.assertRegex(
            str(form.fields["metadata_only_retention_period"].help_text),
            r".+ \(in days\)$",
        )

    def test_check_name(self) -> None:
        form = self.form()
        form.check_name(
            self.scenario.workspace, "foo", CollectionCategory.SUITE
        )
        self.assertEqual(form.errors, {})

    def test_check_name_avoid_name_conflict(self) -> None:
        form = self.form()
        form.check_name(
            self.scenario.workspace, "sid", CollectionCategory.SUITE
        )
        self.assertEqual(
            form.errors,
            {
                "name": [
                    "A collection called sid@debian:suite"
                    f" already exists in workspace {self.scenario.workspace}",
                ]
            },
        )

    def test_check_name_avoid_name_conflict_spares_instance(self) -> None:
        form = self.form()
        form.check_name(
            self.scenario.workspace,
            "sid",
            CollectionCategory.SUITE,
            instance=self.sid,
        )
        self.assertEqual(form.errors, {})

    def test_check_name_wrong_name_for_singleton(self) -> None:
        form = self.form()
        form.check_name(
            self.scenario.workspace,
            "wrong",
            CollectionCategory.PACKAGE_BUILD_LOGS,
        )
        self.assertEqual(
            form.errors,
            {"name": ['A singleton collection needs to be named "_"']},
        )

    def test_check_name_wrong_name_for_non_singleton(self) -> None:
        for name in "_", "_wrong":
            with self.subTest(name=name):
                form = self.form()
                form.check_name(
                    self.scenario.workspace, name, CollectionCategory.SUITE
                )
                self.assertEqual(
                    form.errors,
                    {
                        "name": [
                            "The name of a non-singleton collection"
                            ' cannot start with "_"'
                        ]
                    },
                )


class CollectionCreateFormTests(TestCase):
    """Tests for :py:class:`CollectionCreateForm`."""

    scenario = scenarios.DefaultContext(set_current=True)

    sid: ClassVar[Collection]
    default_fields: ClassVar[dict[str, Any]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        cls.default_fields = {
            "full_history_retention_period": 0,
            "metadata_only_retention_period": 0,
            "data": {},
        }

    def form(self, **kwargs: Any) -> CollectionCreateForm:
        """Instantiate the form."""
        return CollectionCreateForm(data={**self.default_fields, **kwargs})

    def test_clean_without_category(self) -> None:
        form = self.form(name="foo", category="invalid")
        with mock.patch.object(form, "check_name") as check_name:
            self.assertFalse(form.is_valid())
        check_name.assert_not_called()

    def test_clean_checks_name(self) -> None:
        form = self.form(name="foo", category=CollectionCategory.SUITE)
        with mock.patch.object(form, "check_name") as check_name:
            self.assertTrue(form.is_valid())
        check_name.assert_called_once_with(
            self.scenario.workspace, "foo", CollectionCategory.SUITE
        )

    def test_save(self) -> None:
        form = self.form(name="foo", category=CollectionCategory.SUITE)
        self.assertTrue(form.is_valid())
        instance = form.save()
        assert isinstance(instance.pk, int)
        self.assertEqual(instance.workspace, self.scenario.workspace)
        self.assertEqual(
            instance.retains_artifacts, Collection.RetainsArtifacts.ALWAYS
        )

    def test_save_commit_false(self) -> None:
        form = self.form(name="foo", category=CollectionCategory.SUITE)
        self.assertTrue(form.is_valid())
        instance = form.save(commit=False)
        self.assertIsNone(instance.pk)


class CollectionUpdateFormTests(TestCase):
    """Tests for :py:class:`CollectionUpdateForm`."""

    scenario = scenarios.DefaultContext(set_current=True)

    sid: ClassVar[Collection]
    experimental: ClassVar[Collection]
    default_fields: ClassVar[dict[str, Any]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        cls.experimental = cls.playground.create_collection(
            "experimental", CollectionCategory.SUITE
        )
        cls.default_fields = {
            "full_history_retention_period": 0,
            "metadata_only_retention_period": 0,
            "data": {},
        }

    def form(self, instance: Collection, **kwargs: Any) -> CollectionUpdateForm:
        """Instantiate the form."""
        return CollectionUpdateForm(
            data={**self.default_fields, **kwargs}, instance=instance
        )

    def test_clean_checks_name(self) -> None:
        form = self.form(
            self.experimental, name="foo", category=CollectionCategory.SUITE
        )
        with mock.patch.object(form, "check_name") as check_name:
            self.assertTrue(form.is_valid())
        check_name.assert_called_once_with(
            self.scenario.workspace,
            "foo",
            CollectionCategory.SUITE,
            instance=self.experimental,
        )


class CollectionRelationCreateFormTests(TestCase):
    """Tests for CollectionRelationCreateForm."""

    trixie: ClassVar[Collection]
    trixie_backports: ClassVar[Collection]
    personal: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.trixie = cls.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        cls.trixie_backports = cls.playground.create_collection(
            "trixie-backports", CollectionCategory.SUITE
        )
        cls.personal = cls.playground.create_collection(
            "personal", CollectionCategory.SUITE
        )

    def test_form_all_types(self) -> None:
        form = CollectionRelationCreateForm(
            source_collection=self.personal,
            allowed_types=set(CollectionRelation.Types),
        )

        self.assertCountEqual(form.fields.keys(), ["type", "target"])
        type_field = form.fields["type"]

        assert isinstance(type_field, ChoiceField)
        assert isinstance(type_field.choices, Iterable)

        self.assertEqual(
            type_field.choices,
            [
                ("suite-based-on", "Based on"),
                ("suite-default-qa-results", "Default QA results"),
                ("suite-forked-from", "Forked from"),
                ("suite-requires", "Requires"),
                ("suite-targeting", "Targeting"),
            ],
        )

        target_field = form.fields["target"]

        assert hasattr(target_field, "queryset")
        self.assertTrue(target_field.queryset.ordered)

    def test_form_no_types(self) -> None:
        form = CollectionRelationCreateForm(
            source_collection=self.personal,
            allowed_types=set(),
        )

        type_field = form.fields["type"]

        assert isinstance(type_field, ChoiceField)
        assert isinstance(type_field.choices, Iterable)

        self.assertEqual(type_field.choices, [])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utilities used by debusine.project."""

import os

from django.core.exceptions import ImproperlyConfigured


def read_secret_key(secret_key_file: str | os.PathLike[str]) -> str:
    """
    Return the content of secret_key_file.

    If secret_key_file cannot be read, it raises ImproperlyConfigured.
    """
    try:
        if bool(os.stat(secret_key_file).st_mode & 0o077):
            raise ImproperlyConfigured(
                f'Permission too open for {secret_key_file}. '
                'Make sure that the file is not accessible by '
                'group or others'
            )
        with open(secret_key_file) as f:
            return f.read().strip()
    except OSError as exc:
        raise ImproperlyConfigured(f"Cannot read {secret_key_file}: {exc}")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the UI helper infrastructure."""

from typing import Any, cast, override
from unittest import mock

from django.db.models import Model
from django.http import HttpRequest
from django.test import RequestFactory

from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.ui.base import UI, _helper_registry


def make_mock_model(label: str = "test") -> type[Model]:
    """Create a mock model class."""

    class _MockModel:
        _meta = mock.Mock()
        _meta.label_lower = label

        def __init__(self, pk: int = 1) -> None:
            self.pk = pk
            self.can_always_args: Any = None

        def can_always(self, *args: Any, **kwargs: Any) -> bool:
            self.can_always_args = (args, kwargs)
            return True

    return cast(type[Model], _MockModel)


def make_helper(model_class: type[Model]) -> type[UI[Model]]:
    """Create a mock helper for a model class."""

    class Helper(UI[model_class]):  # type: ignore[valid-type]
        pass

    return Helper


class UITests(TestCase):
    """Test the UI base class."""

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    def setUp(self) -> None:
        super().setUp()
        self.enterContext(UI.preserve_registry())

    def request(self) -> HttpRequest:
        factory = RequestFactory()
        request = factory.get("/")
        return request

    def test_helper_registry(self) -> None:
        MockModel = make_mock_model()
        Helper = make_helper(MockModel)

        self.assertIs(_helper_registry["test"], Helper)

    def test_only_one_helper_per_model(self) -> None:
        MockModel = make_mock_model()
        Helper = make_helper(MockModel)
        with self.assertRaisesRegex(
            AssertionError,
            f"{Helper} already registered as helper for {MockModel}",
        ):
            make_helper(MockModel)

    def test_get(self) -> None:
        request = self.request()
        MockModel = make_mock_model()
        Helper = make_helper(MockModel)

        instance = MockModel(pk=1)
        helper = Helper.get(request, instance)
        self.assertIsInstance(helper, Helper)
        self.assertIs(helper.request, request)
        self.assertIs(helper.instance, instance)

        # Instances are reused
        self.assertIs(Helper.get(request, MockModel(pk=1)), helper)

        # An instance with a different pk gets a different helper instance
        instance2 = MockModel(pk=2)
        helper2 = Helper.get(request, instance2)
        self.assertIsInstance(helper2, Helper)
        self.assertIs(helper2.request, request)
        self.assertIs(helper2.instance, instance2)
        self.assertIs(Helper.get(request, MockModel(pk=2)), helper2)

    def test_for_instance(self) -> None:
        request = self.request()

        FooModel = make_mock_model("foo")
        FooHelper = make_helper(FooModel)
        BarModel = make_mock_model("bar")
        BarHelper = make_helper(BarModel)

        foo = FooModel()
        foo_helper = UI.for_instance(request, foo)
        self.assertIsInstance(foo_helper, FooHelper)
        self.assertIs(UI.for_instance(request, FooModel()), foo_helper)

        bar = BarModel()
        bar_helper = UI.for_instance(request, bar)
        self.assertIsInstance(bar_helper, BarHelper)
        self.assertIs(UI.for_instance(request, BarModel()), bar_helper)

    def test_request_isolation(self) -> None:
        """UI helpers are request specific."""
        requestA = self.request()
        requestB = self.request()
        MockModel = make_mock_model()
        Helper = make_helper(MockModel)

        instance = MockModel(pk=1)

        helperA = Helper.get(requestA, instance)
        helperB = Helper.get(requestB, instance)

        self.assertIsNot(helperA, helperB)
        self.assertIs(Helper.get(requestA, instance), helperA)
        self.assertIs(Helper.get(requestB, instance), helperB)

    def test_predicate_proxy(self) -> None:
        request = self.request()
        MockModel = make_mock_model()
        make_helper(MockModel)
        instance = MockModel(pk=1)
        helper = UI.for_instance(request, instance)
        self.assertTrue(helper.can_always)
        assert hasattr(instance, "can_always_args")
        self.assertEqual(
            instance.can_always_args, ((self.scenario.context_ui(),), {})
        )

    def test_predicate_proxy_missing(self) -> None:
        request = self.request()
        MockModel = make_mock_model()
        make_helper(MockModel)
        instance = MockModel(pk=1)
        helper = UI.for_instance(request, instance)
        with self.assertRaisesRegex(
            AttributeError,
            r"'_MockModel' object has no attribute 'can_invalid'",
        ):
            helper.can_invalid

    def test_attribute_error(self) -> None:
        request = self.request()
        MockModel = make_mock_model()
        make_helper(MockModel)
        instance = MockModel(pk=1)
        helper = UI.for_instance(request, instance)
        with self.assertRaisesRegex(
            AttributeError, r"'Helper' object has no attribute 'invalid_member'"
        ):
            helper.invalid_member



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Confirm task."""

from django.core.exceptions import ValidationError

from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.server.tasks.wait import Confirm
from debusine.server.tasks.wait.models import ConfirmData, DenyAction
from debusine.server.tasks.wait.tests.helpers import WaitTaskTestCase
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.models import Confirmation, OutputData


class ConfirmTaskTests(WaitTaskTestCase[Confirm]):
    """Test the Confirm task."""

    scenario = scenarios.DefaultScopeUser()

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_no_additional_properties(self) -> None:
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(ValidationError, error_msg):
            self.playground.create_wait_task(
                task_name="confirm",
                task_data={"input": {}},
                workflow_data=WorkRequestWorkflowData(needs_input=True),
            )

    def test_label(self) -> None:
        self.assertEqual(self.task.get_label(), "wait for confirmation")

    def test_respond_confirm(self) -> None:
        self.mark_running = True
        self.task.respond(user=self.scenario.user, confirm=True)

        work_request = self.work_request
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
        self.assertEqual(work_request.completed_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(confirmation=Confirmation(confirmed=True)),
        )

    def test_respond_deny_fail(self) -> None:
        self.mark_running = True
        self.task_data["deny_action"] = DenyAction.FAIL
        self.task.respond(
            user=self.scenario.user, confirm=False, comment="Broken"
        )

        work_request = self.work_request
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.FAILURE)
        self.assertEqual(work_request.completed_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                confirmation=Confirmation(confirmed=False, comment="Broken")
            ),
        )

    def test_respond_deny_abort(self) -> None:
        self.scenario.set_current()
        workflow = self.playground.create_workflow(mark_running=True)
        work_request = workflow.create_child_wait(
            task_name="confirm",
            task_data=ConfirmData(deny_action=DenyAction.ABORT),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        workflow.create_child_worker(task_name="noop")
        self.playground.advance_work_request(work_request, mark_running=True)
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        task.respond(user=self.scenario.user, confirm=False, comment="Broken")

        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.ABORTED)
        self.assertEqual(work_request.aborted_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                confirmation=Confirmation(confirmed=False, comment="Broken")
            ),
        )
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.RUNNING)

    def test_respond_deny_abort_workflow(self) -> None:
        self.scenario.set_current()
        workflow = self.playground.create_workflow(mark_running=True)
        work_request = workflow.create_child_wait(
            task_name="confirm",
            task_data=ConfirmData(deny_action=DenyAction.ABORT_WORKFLOW),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        workflow.create_child_worker(task_name="noop")
        self.playground.advance_work_request(work_request, mark_running=True)
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        task.respond(user=self.scenario.user, confirm=False, comment="Broken")

        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.ABORTED)
        self.assertEqual(work_request.aborted_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                confirmation=Confirmation(confirmed=False, comment="Broken")
            ),
        )
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.ABORTED)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View for DebDiff work request."""

import re
from collections import defaultdict
from enum import StrEnum
from typing import Any, TypedDict, override

from unidiff import PatchSet

from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianUpload,
    TaskTypes,
    get_source_package_name,
    get_source_package_version,
)
from debusine.db.models import Artifact, ArtifactRelation, FileInArtifact
from debusine.tasks import DebDiff
from debusine.web.views import view_utils
from debusine.web.views.artifacts import ArtifactPlugin
from debusine.web.views.files import FileWidget
from debusine.web.views.work_request import WorkRequestPlugin


class OperationEnum(StrEnum):
    """Operations that are done in a file in the patch."""

    ADDED = "added"
    REMOVED = "removed"
    MODIFIED = "modified"


class DebDiffFileSummary(TypedDict):
    """Summary of a file-level change in a source debdiff."""

    path: str
    diff_line_number: int
    operation: OperationEnum


class DebDiffSection(StrEnum):
    """Sections of a debdiff output."""

    ADDED_FILES = "added_files"
    REMOVED_FILES = "removed_files"
    CONTROL_DIFFS = "control_diffs"
    CONTROL_DIFFS_PACKAGES = "control_diffs_packages"
    CONTROL_DIFFS_PACKAGES_NO_CHANGES = "control_diffs_packages_no_changes"


class DebDiffBinaryParseContext(TypedDict):
    """Defines the result structure of a parsed debdiff binary section."""

    added_files: list[str]
    removed_files: list[str]
    control_diffs: list[str]
    control_diffs_packages: dict[str, list[str]]
    control_diffs_packages_no_changes: list[str]


class DebDiffBinaryParsed(TypedDict):
    """Defines the result structure as it's going to be used by the template."""

    added_files: str
    removed_files: str
    control_diffs: str
    control_diffs_packages: dict[str, str]
    control_diffs_packages_items: list[tuple[str, str]]
    control_diffs_packages_no_changes: list[str]


class DebDiffViewArtifactPlugin(ArtifactPlugin):
    """View for DebDiff Artifact."""

    artifact_category = ArtifactCategory.DEBDIFF
    template_name = "web/artifact-detail.html"
    name = "debdiff"

    object_name = "artifact"

    @classmethod
    def _summarise_debdiff_source(
        cls, debdiff: str
    ) -> list[DebDiffFileSummary]:
        patch = PatchSet(debdiff)

        summary: list[DebDiffFileSummary] = []

        lines = None
        last_binary_line = 0

        for file in patch:
            if file.is_added_file:
                operation = OperationEnum.ADDED
            elif file.is_removed_file:
                operation = OperationEnum.REMOVED
            elif file.is_modified_file:
                operation = OperationEnum.MODIFIED
            else:
                raise ValueError(f"Unexpected file status in diff: {file.path}")

            if file.is_binary_file:
                # find the line in the debdiff report of this binary change
                # Unidiff is not giving this information at the moment
                # See https://github.com/matiasb/python-unidiff/issues/122
                if lines is None:
                    # Avoid splitting for each binary file
                    lines = debdiff.splitlines()

                # Always true in practice, but the type system doesn't know
                # that.
                assert file.patch_info is not None

                diff_line_number = (
                    lines.index(file.patch_info[0].rstrip(), last_binary_line)
                    + 1
                )
                # keep searching from previous one
                last_binary_line = diff_line_number
                summary.append(
                    {
                        "path": file.path,
                        "diff_line_number": diff_line_number,
                        "operation": operation,
                    }
                )
            else:
                # Always true in practice, but the type system doesn't know
                # that.
                assert file[0][0].diff_line_no is not None

                # first diff line number of the first hunk of the file
                #
                # subtracts 3: skip back over the one-line hunk header (@@)
                # and the two-line file header (--- and +++)
                diff_line_number = file[0][0].diff_line_no - 3

                summary.append(
                    {
                        "path": file.path,
                        "diff_line_number": diff_line_number,
                        "operation": operation,
                    }
                )

        return summary

    @staticmethod
    def _detect_section_package(
        line: str,
    ) -> tuple[DebDiffSection | None, str | None]:
        """Detect if a line is the start of a new debdiff section or package."""
        control_files_re = re.compile(r"^Control files of package (\S+):")
        no_differences_re = re.compile(
            r"^No differences were encountered between "
            r"the control files of package (\S+)$"
        )

        if line in (
            "Files in second .changes but not in first",
            "Files in second .deb but not in first",
            "Files in second set of .debs but not in first",
        ):
            return DebDiffSection.ADDED_FILES, None

        elif line in (
            "Files in first .changes but not in second",
            "Files in first .deb but not in second",
            "Files in first set of .debs but not in second",
        ):
            return DebDiffSection.REMOVED_FILES, None

        elif line.startswith("Control files: lines which differ"):
            return DebDiffSection.CONTROL_DIFFS, None

        elif m := control_files_re.match(line):
            return DebDiffSection.CONTROL_DIFFS_PACKAGES, m.group(1)

        elif m := no_differences_re.match(line):
            return DebDiffSection.CONTROL_DIFFS_PACKAGES_NO_CHANGES, m.group(1)

        return None, None

    @staticmethod
    def _append_to_result(
        result: DebDiffBinaryParseContext,
        section: DebDiffSection,
        package: str | None,
        line: str,
    ) -> None:
        value = result[section.value]

        match value:
            case list():
                value.append(line)
            case dict():
                subsection = value

                # If it's a dict we have a package and we have information for
                # each package to add
                assert package is not None

                subsection[package].append(line)

            case _ as unreachable:
                raise NotImplementedError(f"{unreachable!r} not implemented")

    @classmethod
    def _parse_debdiff_binary(cls, debdiff: str) -> DebDiffBinaryParsed:
        result: DebDiffBinaryParseContext = {
            DebDiffSection.ADDED_FILES.value: [],
            DebDiffSection.REMOVED_FILES.value: [],
            DebDiffSection.CONTROL_DIFFS.value: [],
            DebDiffSection.CONTROL_DIFFS_PACKAGES.value: defaultdict(list),
            DebDiffSection.CONTROL_DIFFS_PACKAGES_NO_CHANGES.value: [],
        }

        current_section: DebDiffSection | None = None
        current_package: str | None = None
        has_data = False

        for line in debdiff.splitlines():
            line = line.rstrip()

            if set(line) == {"-"}:
                # Line is a header underline
                continue

            if line == "":
                current_section = None
                current_package = None
                continue

            if (
                line
                == "No differences were encountered between the control files"
            ):
                has_data = True
                break

            section, package = cls._detect_section_package(line)

            has_data |= section is not None

            if section == DebDiffSection.CONTROL_DIFFS_PACKAGES_NO_CHANGES:
                assert package is not None
                result[
                    DebDiffSection.CONTROL_DIFFS_PACKAGES_NO_CHANGES.value
                ].append(package)

            if section is not None:
                current_section = section
                current_package = package
                continue

            if current_section is None and has_data:
                raise ValueError(f"Failed to parse line: {line!r}")

            if current_section is not None:
                cls._append_to_result(
                    result, current_section, current_package, line
                )

        if not has_data:
            raise ValueError("Cannot parse any information from debdiff output")

        return cls._render_binary_parsed(result)

    @override
    def get_context_data(self) -> dict[str, Any]:
        slug = "debdiff"

        specialized_tab = {
            "specialized_tab": {
                "label": "DebDiff",
                "slug": slug,
                "template": "web/_debdiff-artifact-detail.html",
            }
        }

        try:
            debdiff_txt_file_in_artifact = self.artifact.fileinartifact_set.get(
                path=DebDiff.CAPTURE_OUTPUT_FILENAME
            )
        except FileInArtifact.DoesNotExist:
            return {
                **specialized_tab,
                "debdiff_artifact_error": "Artifact does not contain "
                f"{DebDiff.CAPTURE_OUTPUT_FILENAME} ",
            }

        try:
            debdiff_bytes = view_utils.read_file(
                debdiff_txt_file_in_artifact, max_file_size=50 * 1024 * 1024
            )
        except view_utils.FileInArtifactError as exc:
            return {
                **specialized_tab,
                "debdiff_artifact_error": exc.to_html(),
            }

        debdiff_text = debdiff_bytes.decode(errors="replace")

        debdiff_absolute_path = debdiff_txt_file_in_artifact.get_absolute_url()

        related_to = self.artifact.relations.filter(
            type=ArtifactRelation.Relations.RELATES_TO
        ).first()

        if related_to is None:
            return {
                **specialized_tab,
                "debdiff_artifact_error": "Artifact is not complete "
                "(missing 'relates-to' relation)",
            }

        if related_to.target.category == ArtifactCategory.SOURCE_PACKAGE:
            # The original and new artifacts are source artifacts
            debdiff_parser_error = None
            try:
                debdiff_source_summary = self._summarise_debdiff_source(
                    debdiff_text
                )
            except ValueError as exc:
                debdiff_source_summary = None
                debdiff_parser_error = str(exc)

            debdiff_source_file_widget = FileWidget.create(
                debdiff_txt_file_in_artifact,
                file_tag=slug,
                file_contents=debdiff_bytes,
            )

            return {
                **specialized_tab,
                "debdiff_artifact_url": debdiff_absolute_path,
                "debdiff_source_summary": debdiff_source_summary,
                "debdiff_source_file_widget": debdiff_source_file_widget,
                "debdiff_parser_error": debdiff_parser_error,
                "debdiff_differences_reported": debdiff_source_summary
                is not None
                and len(debdiff_source_summary) != 0,
                "debdiff_file_tag": slug,
            }

        else:
            # The original and new artifacts were binary artifacts
            debdiff_parser_error = None
            try:
                debdiff_binary_parsed = self._parse_debdiff_binary(debdiff_text)

            except ValueError as exc:
                debdiff_binary_parsed = None
                debdiff_parser_error = str(exc)

            return {
                **specialized_tab,
                "debdiff_artifact_url": debdiff_absolute_path,
                "debdiff_binary_parsed": debdiff_binary_parsed,
                "debdiff_parser_error": debdiff_parser_error,
                "debdiff_differences_reported": (
                    debdiff_binary_parsed
                    and any(
                        debdiff_binary_parsed[section.value]
                        for section in DebDiffSection
                    )
                ),
            }

    @staticmethod
    def _render_binary_parsed(
        parsed: DebDiffBinaryParseContext,
    ) -> DebDiffBinaryParsed:
        """Convert lists to template-friendly strings."""
        rendered_packages = {
            pkg: "\n".join(lines)
            for pkg, lines in parsed["control_diffs_packages"].items()
        }
        return {
            "added_files": "\n".join(parsed["added_files"]),
            "removed_files": "\n".join(parsed["removed_files"]),
            "control_diffs": "\n".join(parsed["control_diffs"]),
            "control_diffs_packages": rendered_packages,
            "control_diffs_packages_items": list(rendered_packages.items()),
            "control_diffs_packages_no_changes": list(
                parsed["control_diffs_packages_no_changes"]
            ),
        }


class DebDiffViewWorkRequestPlugin(WorkRequestPlugin):
    """View for DebDiff work request."""

    task_type = TaskTypes.WORKER
    task_name = "debdiff"

    @override
    def do_get_description_data(self) -> dict[str, Any]:
        dynamic_data = self.task.dynamic_data
        assert dynamic_data is not None

        data: dict[str, Any] = {}

        if (
            source_artifacts := dynamic_data.input_source_artifacts_ids
        ) is not None:
            data["source_artifact_original_id"] = source_artifacts[0]
            data["source_artifact_new_id"] = source_artifacts[1]

        elif (
            binary_artifacts := dynamic_data.input_binary_artifacts_ids
        ) is not None:
            source_package_names: set[str] = set()

            for binary_artifact in Artifact.objects.filter(
                id__in=binary_artifacts[0]
            ):
                artifact_data = binary_artifact.create_data()

                assert isinstance(
                    artifact_data,
                    (DebianBinaryPackage, DebianUpload),
                )
                src_pkg_name = get_source_package_name(artifact_data)
                src_pkg_version = get_source_package_version(artifact_data)
                source_package_names.add(f"{src_pkg_name}_{src_pkg_version}")

            data["source_package_names"] = sorted(source_package_names)
            data["binary_artifacts_new_count"] = len(binary_artifacts[1])
            data["binary_artifacts_architecture"] = (
                self.task.data.build_architecture
            )
        else:
            # This branch is for WorkRequests where
            # dynamic data is quite incomplete (old WorkRequests)
            data["parameter_summary"] = dynamic_data.parameter_summary

        data["environment_id"] = dynamic_data.environment_id
        data["extra_arguments"] = self.task.data.extra_flags

        return data

    @override
    def get_context_data(self) -> dict[str, Any]:
        return {
            "description_template": "web/_debdiff-description.html",
            "description_data": self.get_description_data(),
        }



# Generated by Django 4.2.21 on 2025-09-17 12:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0017_assetusage_db_assetusage_unique_asset_workspace'),
    ]

    operations = [
        migrations.AlterField(
            model_name='assetusagerole',
            name='role',
            field=models.CharField(
                choices=[
                    ('signer', 'Signer'),
                    ('repository_signer', 'Repository signer'),
                ],
                max_length=32,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for uploading Debian packages to an upload queue."""

import logging
import os
from pathlib import Path
from typing import ClassVar, override
from unittest import mock

from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.db.models import Artifact, File, Workspace
from debusine.server.tasks import PackageUpload
from debusine.server.tasks.models import PackageUploadDynamicData
from debusine.server.tasks.tests.ftp_server import FakeFTPServerThread
from debusine.server.tasks.tests.helpers import ServerTaskTestCase
from debusine.server.tasks.tests.sftp_server import FakeSFTPServerThread


class PackageUploadTests(ServerTaskTestCase[PackageUpload]):
    """Tests for :py:class:`PackageUpload`."""

    task: PackageUpload
    upload: ClassVar[Artifact]

    SAMPLE_TASK_DATA = {
        "target": "ftp://upload.example.org/queue/",
    }

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        filenames = [
            "foo_1.0.orig.tar.xz",
            "foo_1.0-1.debian.tar.xz",
            "foo_1.0-1.dsc",
        ]
        [cls.upload, _] = cls.playground.create_artifact(
            paths=["foo_1.0-1_source.changes", *filenames],
            category=ArtifactCategory.UPLOAD,
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": "source",
                    "Files": [{"name": filename} for filename in filenames],
                },
            },
            create_files=True,
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["input"] = {"upload": self.upload.id}

        # Silence logging from pyftpdlib.
        pyftpdlib_logger = logging.getLogger("pyftpdlib")
        self.addCleanup(setattr, pyftpdlib_logger, "handlers", [])
        pyftpdlib_logger.addHandler(logging.NullHandler())

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task, strip_task_inputs=True),
            PackageUploadDynamicData(input_upload_id=self.upload.id),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.task.dynamic_data = None
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = PackageUploadDynamicData(input_upload_id=1)
        self.assertEqual(self.task.get_input_artifacts_ids(), [1])

    def test_fetch_upload_wrong_category(self) -> None:
        """fetch_upload checks the category of the input.upload artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        self.task_data["input"] = {"upload": artifact.id}

        self.assertIsNone(
            self.task.fetch_upload(self.create_temporary_directory())
        )

        assert self.task._debug_log_files_directory is not None
        self.assertEqual(
            Path(
                self.task._debug_log_files_directory.name, "fetch_upload.log"
            ).read_text(),
            "Expected input.upload of category debian:upload; got "
            "debusine:test\n",
        )

    def test_fetch_upload_gathers_upload_paths(self) -> None:
        """fetch_upload gathers information from the input.upload artifact."""
        destination = self.create_temporary_directory()
        expected_upload_hashes = {
            destination / fia.path: fia.file.hash_digest
            for fia in self.upload.fileinartifact_set.select_related("file")
        }

        upload_paths = self.task.fetch_upload(destination)

        assert upload_paths is not None
        # The files are downloaded with the correct contents.
        self.assertEqual(
            {path: File.calculate_hash(path) for path in upload_paths},
            expected_upload_hashes,
        )
        # The .changes file is sorted to the end of upload_paths.
        self.assertTrue(upload_paths[-1].name.endswith(".changes"))

    def test_execute_no_upload_paths(self) -> None:
        """execute() returns FAILURE if fetch_upload returned no paths."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        self.task_data["input"] = {"upload": artifact.id}
        self.task.dynamic_data = PackageUploadDynamicData(
            input_upload_id=artifact.id
        )
        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )

        self.assertEqual(self.task.execute(), WorkRequestResults.FAILURE)

    def test_execute_ftp(self) -> None:
        """execute() can make an upload using FTP."""
        username = "user"
        remote_home_directory = self.create_temporary_directory()
        (remote_home_directory / "queue").mkdir()
        server = FakeFTPServerThread(username, remote_home_directory)
        server.start()
        self.addCleanup(server.stop)
        host, port = server.address
        self.task_data["target"] = f"ftp://{username}@{host}:{port}/queue/"
        directory = self.create_temporary_directory()
        deb_path = directory / "foo_1.0-1_amd64.deb"
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_deb_file(deb_path)
        self.write_changes_file(changes_path, [deb_path])

        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )

        with mock.patch.object(
            self.task, "fetch_upload", return_value=[deb_path, changes_path]
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        for path in (deb_path, changes_path):
            self.assertEqual(
                (remote_home_directory / "queue" / path.name).read_bytes(),
                path.read_bytes(),
            )

    def test_execute_ftp_no_path(self) -> None:
        """execute() handles FTP URLs with no path."""
        username = "user"
        remote_home_directory = self.create_temporary_directory()
        server = FakeFTPServerThread(username, remote_home_directory)
        server.start()
        self.addCleanup(server.stop)
        host, port = server.address
        self.task_data["target"] = f"ftp://{username}@{host}:{port}"
        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )
        directory = self.create_temporary_directory()
        deb_path = directory / "foo_1.0-1_amd64.deb"
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_deb_file(deb_path)
        self.write_changes_file(changes_path, [deb_path])

        with mock.patch.object(
            self.task, "fetch_upload", return_value=[deb_path, changes_path]
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        for path in (deb_path, changes_path):
            self.assertEqual(
                (remote_home_directory / path.name).read_bytes(),
                path.read_bytes(),
            )

    def test_execute_ftp_delayed(self) -> None:
        """execute() can make a delayed upload using FTP."""
        username = "user"
        remote_home_directory = self.create_temporary_directory()
        (remote_home_directory / "queue" / "DELAYED" / "3-day").mkdir(
            parents=True
        )
        server = FakeFTPServerThread(username, remote_home_directory)
        server.start()
        self.addCleanup(server.stop)
        host, port = server.address
        self.task_data["target"] = f"ftp://{username}@{host}:{port}/queue/"
        self.task_data["delayed_days"] = 3
        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )
        directory = self.create_temporary_directory()
        deb_path = directory / "foo_1.0-1_amd64.deb"
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_deb_file(deb_path)
        self.write_changes_file(changes_path, [deb_path])

        with mock.patch.object(
            self.task, "fetch_upload", return_value=[deb_path, changes_path]
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        for path in (deb_path, changes_path):
            self.assertEqual(
                (
                    remote_home_directory
                    / "queue"
                    / "DELAYED"
                    / "3-day"
                    / path.name
                ).read_bytes(),
                path.read_bytes(),
            )

    def test_execute_ftp_retries_timeouts(self) -> None:
        """execute() retries FTP upload timeouts."""
        directory = self.create_temporary_directory()
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_changes_file(changes_path, [])

        with (
            mock.patch.object(
                self.task, "fetch_upload", return_value=[changes_path]
            ),
            mock.patch(
                "debusine.server.tasks.package_upload.FTP",
                side_effect=[TimeoutError(f"Boom {i}") for i in range(3)],
            ),
            self.assertRaisesRegex(TimeoutError, "Boom 2"),
        ):
            self.task.execute()

        assert self.task._debug_log_files_directory is not None
        self.assertEqual(
            Path(
                self.task._debug_log_files_directory.name, "package-upload.log"
            ).read_text(),
            "".join(
                [
                    f"Uploading to {self.task.data.target}\n",
                    "Upload failed: Boom 0\n",
                    "Upload failed: Boom 1\n",
                    "Upload failed: Boom 2\n",
                ]
            ),
        )

    def test_execute_sftp(self) -> None:
        """execute() can make an upload using SFTP."""
        queue_directory = self.create_temporary_directory()
        server = FakeSFTPServerThread()
        home_directory = self.create_temporary_directory()
        (home_directory / ".ssh").mkdir()
        (id_ed25519 := home_directory / ".ssh" / "id_ed25519").write_text(
            server.user_key.private_openssh
        )
        (home_directory / ".ssh" / "config").write_text(
            f"IdentityFile {id_ed25519}\n"
        )
        server.start()
        self.addCleanup(server.stop)
        host, port = server.address
        self.task_data["target"] = f"sftp://user@{host}:{port}{queue_directory}"
        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )
        directory = self.create_temporary_directory()
        deb_path = directory / "foo_1.0-1_amd64.deb"
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_deb_file(deb_path)
        self.write_changes_file(changes_path, [deb_path])

        with (
            mock.patch.object(
                self.task, "fetch_upload", return_value=[deb_path, changes_path]
            ),
            mock.patch.dict(os.environ, {"HOME": str(home_directory)}),
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        for path in (deb_path, changes_path):
            self.assertEqual(
                (queue_directory / path.name).read_bytes(), path.read_bytes()
            )

    def test_execute_sftp_delayed(self) -> None:
        """execute() can make a delayed upload using SFTP."""
        queue_directory = self.create_temporary_directory()
        (queue_directory / "DELAYED" / "3-day").mkdir(parents=True)
        server = FakeSFTPServerThread()
        home_directory = self.create_temporary_directory()
        (home_directory / ".ssh").mkdir()
        (id_ed25519 := home_directory / ".ssh" / "id_ed25519").write_text(
            server.user_key.private_openssh
        )
        (home_directory / ".ssh" / "config").write_text(
            f"IdentityFile {id_ed25519}\n"
        )
        server.start()
        self.addCleanup(server.stop)
        host, port = server.address
        self.task_data["target"] = f"sftp://user@{host}:{port}{queue_directory}"
        self.task_data["delayed_days"] = 3
        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )
        directory = self.create_temporary_directory()
        deb_path = directory / "foo_1.0-1_amd64.deb"
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_deb_file(deb_path)
        self.write_changes_file(changes_path, [deb_path])

        with (
            mock.patch.object(
                self.task, "fetch_upload", return_value=[deb_path, changes_path]
            ),
            mock.patch.dict(os.environ, {"HOME": str(home_directory)}),
        ):
            self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        for path in (deb_path, changes_path):
            self.assertEqual(
                (
                    queue_directory / "DELAYED" / "3-day" / path.name
                ).read_bytes(),
                path.read_bytes(),
            )

    def test_execute_sftp_retries_timeouts(self) -> None:
        """execute() retries SFTP upload timeouts."""
        self.task_data["target"] = "sftp://upload.example.org/queue/"
        directory = self.create_temporary_directory()
        changes_path = directory / "foo_1.0-1_amd64.changes"
        self.write_changes_file(changes_path, [])

        with (
            mock.patch.object(
                self.task, "fetch_upload", return_value=[changes_path]
            ),
            mock.patch(
                "debusine.server.tasks.package_upload.Connection",
                side_effect=[TimeoutError(f"Boom {i}") for i in range(3)],
            ),
            self.assertRaisesRegex(TimeoutError, "Boom 2"),
        ):
            self.task.execute()

        assert self.task._debug_log_files_directory is not None
        self.assertEqual(
            Path(
                self.task._debug_log_files_directory.name, "package-upload.log"
            ).read_text(),
            "".join(
                [
                    f"Uploading to {self.task.data.target}\n",
                    "Upload failed: Boom 0\n",
                    "Upload failed: Boom 1\n",
                    "Upload failed: Boom 2\n",
                ]
            ),
        )

    def test_execute_end_to_end(self) -> None:
        """End-to-end execution works."""
        username = "user"
        remote_home_directory = self.create_temporary_directory()
        (remote_home_directory / "queue").mkdir()
        server = FakeFTPServerThread(username, remote_home_directory)
        server.start()
        self.addCleanup(server.stop)
        host, port = server.address
        self.task_data["target"] = f"ftp://{username}@{host}:{port}/queue/"
        self.playground.create_group_role(
            self.task.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.work_request.created_by],
        )
        assert self.task.workspace is not None
        directory = self.create_temporary_directory()
        non_changes_paths = [
            directory / fia.path
            for fia in self.upload.fileinartifact_set.all()
            if not fia.path.endswith(".changes")
        ]
        for path in non_changes_paths:
            path.write_text(path.name)
        [changes_path] = [
            directory / fia.path
            for fia in self.upload.fileinartifact_set.all()
            if fia.path.endswith(".changes")
        ]
        self.write_changes_file(changes_path, non_changes_paths)

        self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        debug_logs_artifact = Artifact.objects.get(
            category=ArtifactCategory.WORK_REQUEST_DEBUG_LOGS,
            created_by_work_request=self.work_request,
        )
        package_upload_log = debug_logs_artifact.fileinartifact_set.get()
        self.assertEqual(package_upload_log.path, "package-upload.log")
        fileobj = package_upload_log.file
        file_backend = self.task.workspace.scope.download_file_backend(fileobj)
        with file_backend.get_stream(fileobj) as file:
            self.assertEqual(
                file.read().decode(),
                "".join(
                    [f"Uploading to {self.task.data.target}\n"]
                    + [
                        f"Uploading {path.name}\n"
                        for path in [*sorted(non_changes_paths), changes_path]
                    ]
                    + ["Upload succeeded\n"]
                ),
            )

    def test_label(self) -> None:
        """Test get_label."""
        self.assertEqual(
            self.task.get_label(), "upload to ftp://upload.example.org/queue/"
        )

    def test_label_delayed(self) -> None:
        """Test get_label for a delayed upload."""
        self.task_data["delayed_days"] = 3

        self.assertEqual(
            self.task.get_label(),
            "upload to ftp://upload.example.org/queue/DELAYED/3-day",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Infrastructure for declarative table definitions."""


class Attribute:
    """
    Base for declarative table attributes, like columns and filters.

    Attributes have a name, which is usually taken from the table member the
    attribute is assigned to. It can optionally be explicitly given in the
    attribute to have names like ``for`` or ``sorted`` which would conflict
    with Python statements or existing names.
    """

    name: str

    def __init__(self, *, name: str | None = None) -> None:
        """Optionally set a table name."""
        if name is not None:
            self.name = name



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""DebusineHttpClient interact with the debusine server API."""

from collections.abc import Callable, Generator
from typing import Any, overload

import requests
from pydantic import BaseModel, ValidationError
from requests.compat import json  # type: ignore[attr-defined]
from requests.structures import CaseInsensitiveDict
from urllib3.util import Retry

from debusine.client import exceptions
from debusine.client.models import PaginatedResponse
from debusine.utils import get_user_agent
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class DebusineHttpClient:
    """DebusineHttpClient interact with the debusine server API."""

    def __init__(
        self,
        api_url: str,
        token: str | None,
        scope: str | None = None,
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> None:
        """
        Initialize DebusineHttpClient.

        :param timeout: timeout for HTTP requests (connect, read).
        """
        self._api_url = api_url
        self._scope = scope
        self._token = token
        self._timeout = timeout

        self._session: requests.Session | None = None

    def get[ClientModel: BaseModel](
        self, path: str, expected_class: type[ClientModel]
    ) -> ClientModel:
        """Make an HTTP GET request to path, return object of expected_class."""
        return self._api_request("GET", self._api_url + path, expected_class)[0]

    def get_single_page_with_headers[ClientModel: BaseModel](
        self, path: str, expected_class: type[ClientModel]
    ) -> tuple[list[ClientModel], CaseInsensitiveDict[str]]:
        """
        Retrieve a single-page listing endpoint and return the response headers.

        Return validated objects of expected_class.
        """
        url = self._api_url + path
        contents, headers = self._api_request("GET", url, PaginatedResponse)

        if contents.next:
            raise exceptions.UnexpectedResponseError(
                f"Expected a single paginated response page from {url}, "
                f"but server returned additional pages"
            )

        try:
            return [
                expected_class.model_validate(raw_object)
                for raw_object in contents.results
            ], headers
        except ValidationError as exc:
            raise exceptions.UnexpectedResponseError(
                f"Server ({url}) did not return valid object. Error: {exc}"
            ) from exc

    def iter_paginated_get[ClientModel: BaseModel](
        self, path: str, expected_class: type[ClientModel]
    ) -> Generator[ClientModel]:
        """
        Iterate through a paginated listing endpoint.

        Yield objects of expected_class.
        """
        url = self._api_url + path
        while True:
            r = self._api_request("GET", url, PaginatedResponse)[0]
            for result in r.results:
                try:
                    yield expected_class.model_validate(result)
                except ValidationError as exc:
                    raise exceptions.UnexpectedResponseError(
                        f"Server ({url}) did not return valid object. "
                        f"Error: {exc}"
                    ) from exc
            if not r.next:
                break
            url = str(r.next)

    @overload
    def post(
        self,
        path: str,
        expected_class: None,
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
    ) -> None: ...

    @overload
    def post[ClientModel: BaseModel](
        self,
        path: str,
        expected_class: type[ClientModel],
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
    ) -> ClientModel: ...

    def post[ClientModel: BaseModel](
        self,
        path: str,
        expected_class: type[ClientModel] | None,
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
    ) -> ClientModel | None:
        """
        Make an HTTP POST request to path.

        Send data and return an object of expected_class.
        """
        return self._api_request(
            "POST",
            self._api_url + path,
            expected_class,
            data,
            expected_statuses=expected_statuses,
        )[0]

    def put[ClientModel: BaseModel](
        self,
        path: str,
        expected_class: type[ClientModel] | None,
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
        if_match: str | None = None,
    ) -> ClientModel | None:
        """
        Make an HTTP PUT request to path.

        Send data and return an object of expected_class.
        """
        return self._api_request(
            "PUT",
            self._api_url + path,
            expected_class,
            data,
            expected_statuses=expected_statuses,
            if_match=if_match,
        )[0]

    @overload
    def patch(
        self,
        path: str,
        expected_class: None,
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
    ) -> None: ...

    @overload
    def patch[ClientModel: BaseModel](
        self,
        path: str,
        expected_class: type[ClientModel],
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
    ) -> ClientModel: ...

    def patch[ClientModel: BaseModel](
        self,
        path: str,
        expected_class: type[ClientModel] | None,
        data: dict[str, Any] | None,
        *,
        expected_statuses: list[int] | None = None,
    ) -> ClientModel | None:
        """
        Make an HTTP PATCH request to path.

        Send data and return an object of expected_class.
        """
        return self._api_request(
            "PATCH",
            self._api_url + path,
            expected_class,
            data,
            expected_statuses=expected_statuses,
        )[0]

    def delete[ClientModel: BaseModel](
        self,
        path: str,
        *,
        expected_class: type[ClientModel] | None = None,
        expected_statuses: list[int] | None = None,
    ) -> ClientModel | None:
        """Make an HTTP DELETE request to path."""
        return self._api_request(
            "DELETE",
            self._api_url + path,
            expected_class=expected_class,
            expected_statuses=expected_statuses,
        )[0]

    def _method(self, method: str) -> Callable[..., requests.Response]:
        if not self._session:
            raise AssertionError("self._session not set")
        method_to_func: dict[str, Callable[..., requests.Response]] = {
            'GET': self._session.get,
            'POST': self._session.post,
            'PUT': self._session.put,
            'PATCH': self._session.patch,
            'DELETE': self._session.delete,
        }

        if method not in method_to_func:
            allowed_methods = ", ".join(method_to_func.keys())
            raise ValueError(f'Method must be one of: {allowed_methods}')

        return method_to_func[method]

    @staticmethod
    def _handle_response_content[ClientModel: BaseModel](
        content: bytes,
        expected_class: type[ClientModel] | None,
        url: str,
    ) -> ClientModel | None:
        if expected_class is None:
            if content != b"":
                raise exceptions.UnexpectedResponseError(
                    f"Server ({url}) expected to return an empty body. "
                    f"Returned:\n{content!r}"
                )
            else:
                return None
        else:
            try:
                return expected_class.model_validate_json(content)
            except ValidationError as exc:
                raise exceptions.UnexpectedResponseError(
                    f"Server ({url}) did not return valid object. Error: {exc}"
                ) from exc

    @overload
    def _api_request(
        self,
        method: str,
        url: str,
        expected_class: None,
        data: dict[str, Any] | None = None,
        *,
        expected_statuses: list[int] | None = None,
        if_match: str | None = None,
    ) -> tuple[None, CaseInsensitiveDict[str]]: ...

    @overload
    def _api_request[ClientModel: BaseModel](
        self,
        method: str,
        url: str,
        expected_class: type[ClientModel],
        data: dict[str, Any] | None = None,
        *,
        expected_statuses: list[int] | None = None,
        if_match: str | None = None,
    ) -> tuple[ClientModel, CaseInsensitiveDict[str]]: ...

    def _api_request[ClientModel: BaseModel](
        self,
        method: str,
        url: str,
        expected_class: type[ClientModel] | None,
        data: dict[str, Any] | None = None,
        *,
        expected_statuses: list[int] | None = None,
        if_match: str | None = None,
    ) -> tuple[ClientModel | None, CaseInsensitiveDict[str]]:
        """
        Request to the server.

        :param method: HTTP method (GET, POST, ...).
        :param url: The complete URL to request.
        :param expected_class: expected object class that the server.
          will return. Used to deserialize the response and return an object.
          If None expects an empty response body.
        :param expected_statuses: defaults to expect HTTP 200. List of HTTP
          status codes that might be return by the call. If it receives
          an unexpected one it raises UnexpectedResponseError.
        :param if_match: if not None, set it to the header "If-Match"
        :raises exceptions.UnexpectedResponseError: the server didn't return
          a valid JSON or returned an unexpected HTTP status code.
        :raises exceptions.WorkRequestNotFound: the server could not find the
          work request.
        :raises exceptions.ClientConnectionError: the client could not connect
          to the server.
        :raises ValueError: invalid options passed.
        :raises DebusineError: the server returned an error with
          application/problem+json. Contains the detail message.
        :raises exceptions.ClientForbiddenError: the server returned HTTP 403.
        :return: an object of the expected_class or None if expected_class
          was None.
        """
        self._ensure_session()

        if data is not None and method == 'GET':
            raise ValueError('data argument not allowed with HTTP GET')

        if expected_statuses is None:
            expected_statuses = [requests.codes.ok]

        optional_kwargs: dict[str, Any] = {}

        if data is not None:
            optional_kwargs = {'json': data}

        try:
            headers = {}
            if self._scope is not None:
                headers["X-Debusine-Scope"] = self._scope
            if self._token is not None:
                headers["Token"] = self._token
            if if_match is not None:
                headers["If-Match"] = if_match
            optional_kwargs.setdefault("timeout", self._timeout)

            response = self._method(method)(
                url, headers=headers, **optional_kwargs
            )
        except (requests.exceptions.RequestException, ConnectionError) as exc:
            raise exceptions.ClientConnectionError(
                f'Cannot connect to {url}. Error: {exc}'
            ) from exc

        if error := self._debusine_problem(response):
            raise exceptions.DebusineError(**error)
        elif response.status_code in expected_statuses:
            return self._handle_response_content(
                response.content, expected_class, url
            ), response.headers
        elif response.status_code == requests.codes.not_found:
            raise exceptions.NotFoundError(f'Not found ({url})')
        elif response.status_code == requests.codes.forbidden:
            raise exceptions.ClientForbiddenError(
                f"HTTP 403. Token ({self._token}) is invalid or disabled"
            )
        else:
            raise exceptions.UnexpectedResponseError(
                f'Server returned unexpected status '
                f'code: {response.status_code} ({url})'
            )

    def _ensure_session(self) -> None:
        if self._session is None:
            self._session = requests.Session()
            adapter = requests.adapters.HTTPAdapter(
                max_retries=Retry(
                    total=4,
                    backoff_factor=0.1,
                    status_forcelist=[
                        requests.codes.server_error,  # 500
                        requests.codes.service_unavailable,  # 503
                        requests.codes.gateway_timeout,  # 504
                        # backoff factor will result in a short wait
                        requests.codes.too_many_requests,  # 429
                    ],
                    allowed_methods={"GET", "PATCH", "POST", "PUT"},
                )
            )
            self._session.mount("https://", adapter)
            self._session.headers["User-Agent"] = get_user_agent()

    @staticmethod
    def _debusine_problem(response: requests.Response) -> dict[Any, Any] | None:
        """If response is an application/problem+json returns the body JSON."""
        if response.headers.get("content-type") == "application/problem+json":
            try:
                content = response.json()
            except json.JSONDecodeError:
                return None

            assert isinstance(content, dict)
            return content

        return None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine scope views."""

from typing import Any, override

from django.views.generic import TemplateView

from debusine.db.context import context
from debusine.db.models import Workspace
from debusine.web.views.base import BaseUIView
from debusine.web.views.places import Place


class ScopeDetailView(BaseUIView, TemplateView):
    """Landing page for a scope."""

    # This is anomalous for a detail view in that it does not take arguments,
    # and only displays the current scope. This is due to scopes being the
    # roots of different urlconfs, rather than a parameterized path inside a
    # single urlconf.

    template_name = "web/scope-detail.html"

    @override
    def get_place(self) -> Place:
        return context.require_scope().ui(self.request).place

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Build template context data."""
        ctx = super().get_context_data(**kwargs)
        ctx["workspaces"] = (
            Workspace.objects.in_current_scope()
            .can_display(context.pc)
            .order_by("name")
        )
        return ctx



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for input module."""

import io
import sys
from contextlib import redirect_stderr, redirect_stdout
from typing import Any
from unittest import mock

import yaml

from debusine.test import TestCase
from debusine.utils.input import EditorInteractionError, InvalidData, YamlEditor


class DebusineErrorTest(Exception):
    """Avoid using DebusineError to not import from client."""


class YamlEditorTests(TestCase):
    """Tests for the :py:class:`YamlEditor` class."""

    def _editor(
        self,
        value: dict[str, Any] | None = None,
        help_text: list[str] | None = None,
    ) -> YamlEditor[dict[str, Any]]:
        if value is None:
            value = {}
        editor = YamlEditor(value, help_text=help_text)
        self.addCleanup(editor.cleanup)
        return editor

    def test_init(self) -> None:
        editor = self._editor()
        self.assertEqual(editor.initial, {})
        self.assertEqual(editor.value, {})
        assert editor.editor_file.name.startswith("edit-")

    def test_open_editor(self) -> None:
        editor = self._editor()
        with (
            mock.patch("shutil.which", return_value="foo") as mock_which,
            mock.patch("subprocess.run") as mock_run,
        ):
            editor.open_editor()

        mock_which.assert_called_with("sensible-editor")
        mock_run.assert_called_with(["foo", editor.editor_file.as_posix()])

    def test_open_editor_not_found(self) -> None:
        editor = self._editor()
        with (
            mock.patch("shutil.which", return_value=None),
            self.assertRaisesRegex(
                EditorInteractionError, "sensible-editor not found"
            ),
        ):
            editor.open_editor()

    def test_read_edited(self) -> None:
        test_value = {"test": 1}
        editor = self._editor(test_value)
        editor.editor_file.write_text(yaml.safe_dump(test_value))
        editor.read_edited()
        self.assertEqual(editor.value, test_value)

    def test_read_edited_not_found(self) -> None:
        editor = self._editor()
        editor.editor_file.unlink()
        with self.assertRaisesRegex(
            EditorInteractionError, "Cannot open edited file"
        ):
            editor.read_edited()
        self.assertEqual(editor.value, editor.initial)

    def test_read_edited_empty(self) -> None:
        for data_type in (list, dict):
            with self.subTest(data_type=repr(data_type)):
                editor: YamlEditor[Any] = YamlEditor(data_type())
                editor.value = data_type([("test", 1)])
                editor.editor_file.write_text("")
                editor.read_edited()
                self.assertEqual(editor.value, data_type())

    def test_read_edited_invalid_yaml(self) -> None:
        editor = self._editor()
        editor.editor_file.write_text("{")
        with self.assertRaisesRegex(InvalidData, "Invalid YAML:"):
            editor.read_edited()
        self.assertEqual(editor.value, editor.initial)

    def test_read_edited_invalid_type(self) -> None:
        editor = self._editor()
        editor.editor_file.write_text("[]")
        with self.assertRaisesRegex(
            InvalidData, "Data must be a dict or empty"
        ):
            editor.read_edited()
        self.assertEqual(editor.value, editor.initial)

    def test_message(self) -> None:
        editor = self._editor()
        with (
            redirect_stdout(stdout := io.StringIO()),
            redirect_stderr(stderr := io.StringIO()),
        ):
            editor.message("test message")
        self.assertEqual(stdout.getvalue(), "test message\n")
        self.assertEqual(stderr.getvalue(), "")

    def test_confirm(self) -> None:
        editor = self._editor()
        for reply, expected in (
            ("y", True),
            ("Y", True),
            (" Y ", True),
            ("n", False),
            ("N", False),
            (" N ", False),
        ):
            with (
                self.subTest(reply=reply),
                mock.patch("builtins.input", return_value=reply) as mock_input,
            ):
                self.assertEqual(editor.confirm("prompt"), expected)
                mock_input.assert_called_with("prompt (y/n)> ")

    def test_confirm_retry(self) -> None:
        editor = self._editor()
        with mock.patch(
            "builtins.input", side_effect=["yes", "y"]
        ) as mock_input:
            self.assertTrue(editor.confirm("prompt"))
            mock_input.assert_has_calls([mock.call("prompt (y/n)> ")] * 2)

    def test_help_text(self) -> None:
        def mock_editor(self: YamlEditor[dict[str, Any]]) -> None:
            with self.editor_file.open() as edited:
                lines = list(edited)
            assert lines == ["# line1\n", "# line2\n", "\n", "test: 1\n"]

        editor = self._editor({"test": 1}, help_text=["line1", "line2"])
        with mock.patch(
            "debusine.utils.input.YamlEditor.open_editor",
            side_effect=mock_editor,
            autospec=True,
        ):
            self.assertTrue(editor.edit())
        self.assertEqual(editor.value, {"test": 1})

    def test_edit(self) -> None:
        def write_file(self: YamlEditor[dict[str, Any]]) -> None:
            self.editor_file.write_text("{'test': 1}")

        editor = self._editor()
        with mock.patch(
            "debusine.utils.input.YamlEditor.open_editor",
            side_effect=write_file,
            autospec=True,
        ):
            self.assertTrue(editor.edit())
        self.assertEqual(editor.value, {"test": 1})

    def test_edit_initial_data(self) -> None:
        """The editor is opened in the initial data."""

        def write_file(self_: YamlEditor[dict[str, Any]]) -> None:
            self.assertEqual(self_.editor_file.read_text(), "test: 42\n")

        editor = self._editor({"test": 42})
        with mock.patch(
            "debusine.utils.input.YamlEditor.open_editor",
            side_effect=write_file,
            autospec=True,
        ):
            self.assertTrue(editor.edit())
        self.assertFalse(editor.editor_file.exists())

    def test_edit_fails_user_aborts(self) -> None:
        editor = self._editor({"test": 1})
        with (
            mock.patch(
                "debusine.utils.input.YamlEditor.open_editor",
            ) as mock_open_editor,
            mock.patch(
                "debusine.utils.input.YamlEditor.read_edited",
                side_effect=InvalidData("test"),
            ),
            mock.patch(
                "debusine.utils.input.YamlEditor.message",
            ) as mock_message,
            mock.patch(
                "debusine.utils.input.YamlEditor.confirm", return_value=False
            ),
        ):
            self.assertFalse(editor.edit())
        self.assertEqual(editor.value, {"test": 1})

        mock_open_editor.assert_called_once()
        mock_message.assert_has_calls(
            [
                mock.call("Error reading metadata: test"),
                mock.call(f"Edits left in {editor.editor_file}"),
            ]
        )

        # Check contents of the file
        self.assertTrue(editor.editor_file.exists())
        self.assertEqual(editor.editor_file.read_text(), "test: 1\n")
        editor.editor_file.unlink()

    def test_edit_fails_user_retries(self) -> None:
        call_count = 0

        def mock_read_edited(self_: YamlEditor[dict[str, Any]]) -> bool:
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                raise InvalidData("test")
            else:
                self_.value = {"test": 42}
                return True

        editor = self._editor({"test": 1})
        with (
            mock.patch(
                "debusine.utils.input.YamlEditor.open_editor",
            ) as mock_open_editor,
            mock.patch(
                "debusine.utils.input.YamlEditor.read_edited",
                side_effect=mock_read_edited,
                autospec=True,
            ),
            mock.patch(
                "debusine.utils.input.YamlEditor.confirm", return_value=True
            ),
            mock.patch(
                "debusine.utils.input.YamlEditor.message",
                autospec=True,
            ) as mock_message,
        ):
            self.assertTrue(editor.edit())
        self.assertEqual(editor.value, {"test": 42})

        mock_open_editor.assert_has_calls([mock.call()] * 2)
        self.assertFalse(editor.editor_file.exists())

        mock_message.assert_has_calls(
            [mock.call(mock.ANY, "Error reading metadata: test")]
        )

    def test_edit_apply_loop_success(self) -> None:
        editor = self._editor({"test": 1})
        editor.value = {"test": 42}

        apply = mock.Mock(return_value="result")

        with mock.patch.object(editor, "edit", return_value=True) as edit:
            self.assertEqual(
                editor.edit_apply_loop(
                    apply=apply,
                    apply_exception_class=DebusineErrorTest,
                    print_exception=self.print_exception,
                    content_description="metadata",
                ),
                "result",
            )

        edit.assert_called_once_with()
        apply.assert_called_once_with({"test": 42})

    def test_edit_apply_loop_edit_aborted(self) -> None:
        editor = self._editor({"test": 1})

        apply = mock.Mock()

        with mock.patch.object(editor, "edit", return_value=False) as edit:
            result = editor.edit_apply_loop(
                apply=apply,
                apply_exception_class=DebusineErrorTest,
                print_exception=self.print_exception,
                content_description="metadata",
            )
            self.assertIsNone(result)

        edit.assert_called_once_with()
        apply.assert_not_called()

    def test_edit_apply_loop_no_changes(self) -> None:
        editor = self._editor({"test": 1})
        editor.value = {"test": 1}

        apply = mock.Mock()

        with mock.patch.object(editor, "edit", return_value=True) as edit:
            self.assertIsNone(
                editor.edit_apply_loop(
                    apply=apply,
                    apply_exception_class=DebusineErrorTest,
                    print_exception=self.print_exception,
                    content_description="metadata",
                )
            )

        edit.assert_called_once_with()
        apply.assert_not_called()

    def test_edit_apply_loop_value_error_user_aborts(self) -> None:
        editor = self._editor({"test": 1})
        editor.value = {"test": 42}

        apply = mock.Mock(side_effect=ValueError("invalid value"))

        with (
            mock.patch.object(editor, "edit", return_value=True) as edit,
            mock.patch.object(editor, "confirm", return_value=False) as confirm,
            redirect_stderr(stderr := io.StringIO()),
        ):
            self.assertIsNone(
                editor.edit_apply_loop(
                    apply=apply,
                    apply_exception_class=DebusineErrorTest,
                    print_exception=self.print_exception,
                    content_description="metadata",
                )
            )

        edit.assert_called_once_with()
        apply.assert_called_once_with({"test": 42})
        confirm.assert_called_once_with("\nDo you want to retry the same edit?")
        self.assertEqual(
            stderr.getvalue(),
            "Error parsing metadata\n\ninvalid value\n",
        )

    def test_edit_apply_loop_value_error_user_retries(self) -> None:
        editor = self._editor({"test": 1})

        edited_values = iter(({"test": 42}, {"test": 43}))

        def edit() -> bool:
            editor.value = next(edited_values)
            return True

        apply = mock.Mock(side_effect=[ValueError("invalid value"), "result"])

        with (
            mock.patch.object(editor, "edit", side_effect=edit) as edit_mock,
            mock.patch.object(editor, "confirm", return_value=True) as confirm,
            redirect_stderr(stderr := io.StringIO()),
        ):
            self.assertEqual(
                editor.edit_apply_loop(
                    apply=apply,
                    apply_exception_class=DebusineErrorTest,
                    print_exception=self.print_exception,
                    content_description="metadata",
                ),
                "result",
            )

        self.assertEqual(edit_mock.call_count, 2)
        apply.assert_has_calls(
            [
                mock.call({"test": 42}),
                mock.call({"test": 43}),
            ]
        )
        confirm.assert_called_once_with("\nDo you want to retry the same edit?")
        self.assertEqual(
            stderr.getvalue(),
            "Error parsing metadata\n\ninvalid value\n",
        )

    def test_edit_apply_loop_apply_error_user_aborts(self) -> None:
        editor = self._editor({"test": 1})
        editor.value = {"test": 42}

        error = DebusineErrorTest("Cannot set metadata")
        apply = mock.Mock(side_effect=error)

        with (
            mock.patch.object(editor, "edit", return_value=True) as edit,
            mock.patch.object(editor, "confirm", return_value=False) as confirm,
            redirect_stderr(stderr := io.StringIO()),
        ):
            self.assertIsNone(
                editor.edit_apply_loop(
                    apply=apply,
                    apply_exception_class=DebusineErrorTest,
                    print_exception=self.print_exception,
                    content_description="metadata",
                )
            )

        edit.assert_called_once_with()
        apply.assert_called_once_with({"test": 42})
        confirm.assert_called_once_with("\nDo you want to retry the same edit?")

        self.assertEqual(
            stderr.getvalue(),
            "Error setting metadata:\n\nCannot set metadata\n",
        )

    def test_edit_apply_loop_apply_error_user_retries(self) -> None:
        editor = self._editor({"test": 1})

        edited_values = iter(({"test": 42}, {"test": 43}))

        def edit() -> bool:
            editor.value = next(edited_values)
            return True

        error = DebusineErrorTest("Cannot set metadata")
        apply = mock.Mock(side_effect=[error, "result"])

        with (
            mock.patch.object(editor, "edit", side_effect=edit) as edit_mock,
            mock.patch.object(editor, "confirm", return_value=True) as confirm,
            redirect_stderr(stderr := io.StringIO()),
        ):
            self.assertEqual(
                editor.edit_apply_loop(
                    apply=apply,
                    apply_exception_class=DebusineErrorTest,
                    print_exception=self.print_exception,
                    content_description="metadata",
                ),
                "result",
            )

        self.assertEqual(edit_mock.call_count, 2)
        apply.assert_has_calls(
            [
                mock.call({"test": 42}),
                mock.call({"test": 43}),
            ]
        )
        confirm.assert_called_once_with("\nDo you want to retry the same edit?")
        self.assertIn("Error setting metadata:", stderr.getvalue())
        self.assertIn("Cannot set metadata", stderr.getvalue())

    def test_cleanup(self) -> None:
        editor = self._editor()
        editor.cleanup()
        self.assertFalse(editor.editor_file.exists())

    @staticmethod
    def print_exception(exception: Exception) -> None:
        print(exception, file=sys.stderr)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Implementation of in-memory file backend."""

import contextlib
import io
import logging
import tempfile
from collections import defaultdict
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Self, override

from debusine.db.models import File, FileStore
from debusine.server.file_backend.interface import (
    FileBackendEntryInterface,
    FileBackendInterface,
)
from debusine.server.file_backend.models import MemoryFileBackendConfiguration

logger = logging.getLogger(__name__)


class MemoryFileBackendEntry(
    FileBackendEntryInterface["MemoryFileBackend", str]
):
    """
    An entry in an in-memory file backend.

    In-memory identifiers are the hex digest of the file contents.
    """

    @classmethod
    def _create_from_file(
        cls, backend: "MemoryFileBackend", fileobj: File
    ) -> Self:
        """Make an entry based on a `File`."""
        hash_hex = fileobj.hash_digest.hex()
        return cls(backend=backend, identifier=hash_hex)

    @override
    def __str__(self) -> str:
        """Return a string describing this entry."""
        return self._identifier

    @override
    def get_local_path(self) -> None:
        """Return None: no local paths for in-memory files."""
        return None

    @override
    @contextmanager
    def get_temporary_local_path(self) -> Generator[Path]:
        with tempfile.NamedTemporaryFile(
            prefix="debusine-memory-"
        ) as temp_file:
            temp_file.write(self.backend.storage[self._identifier])
            temp_file.flush()
            yield Path(temp_file.name)

    @override
    def get_url(
        self,
        content_type: str | None = None,  # noqa: ARG002, U100
    ) -> None:
        """Return None: no remote URLs for in-memory files."""
        return None

    @override
    def get_mtime(self) -> None:  # pragma: no cover
        raise NotImplementedError("no last-modified times for in-memory files.")

    @override
    def add(self, local_path: Path) -> None:
        self.backend.storage[File.calculate_hash(local_path).hex()] = (
            local_path.read_bytes()
        )

    @override
    def same_contents(self, local_path: Path) -> bool:
        return self.backend.storage[self._identifier] == local_path.read_bytes()

    @override
    def remove(self) -> None:
        self.backend.storage.pop(self._identifier, None)


class MemoryFileBackend(FileBackendInterface[MemoryFileBackendConfiguration]):
    """In-memory file backend (used for tests)."""

    # In-memory file storages, indexed by configuration names
    storages: dict[str, dict[str, bytes]] = defaultdict(dict)

    def __init__(self, file_store: FileStore) -> None:
        """Initialize MemoryFileBackend."""
        super().__init__()
        self.db_store = file_store
        self.storage = self.storages[self.configuration.name]

    @override
    def get_entry(self, fileobj: File) -> MemoryFileBackendEntry:
        return MemoryFileBackendEntry._create_from_file(self, fileobj)

    def _store_file(self, hash_hex: str, value: bytes) -> None:
        """Add/update a file in the storage."""
        # This is here mainly so it can be more easily mocked than a dict
        # assignment
        self.storage[hash_hex] = value

    @override
    @contextlib.contextmanager
    def get_stream(self, fileobj: File) -> Generator[io.BytesIO]:
        """Yield a file-like object that can be read."""
        hash_hex = fileobj.hash_digest.hex()
        with io.BytesIO(self.storage[hash_hex]) as out:
            yield out

    @override
    def list_entries(self) -> Generator[MemoryFileBackendEntry]:
        for hash_hex in self.storage:
            yield MemoryFileBackendEntry(backend=self, identifier=hash_hex)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utility functions for workflows."""

import functools
from collections.abc import Collection as AbcCollection
from collections.abc import Sequence
from operator import attrgetter
from typing import Any, TYPE_CHECKING, cast, get_args

from django.db.models import QuerySet

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    ArtifactDataWithArch,
    BareDataCategory,
    CollectionCategory,
    DebianBinaryPackage,
    DebianPackageBuildLog,
    DebianSourcePackage,
    DebianSystemTarball,
    DebianUpload,
    DebusinePromise,
    get_architectures,
    get_source_package_name,
)
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.models import Artifact, ArtifactRelation
from debusine.db.models.lookups import LookupResult
from debusine.tasks import TaskConfigError, ensure_artifact_categories
from debusine.tasks.models import (
    BackendType,
    ExtraExternalRepository,
    ExtraRepository,
    LookupMultiple,
    LookupSingle,
)

if TYPE_CHECKING:
    from debusine.server.workflows import Workflow


@functools.lru_cache(maxsize=100)
def source_package(
    workflow: "Workflow[Any, Any]",
    *,
    configuration_key: str = "source_artifact",
) -> Artifact:
    """
    Retrieve the source package artifact from ``configuration_key``.

    If the source artifact is a :artifact:`debian:upload`, returns its
    :artifact:`debian:source-package`.
    """
    artifact = workflow.work_request.lookup_single(
        attrgetter(configuration_key)(workflow.data),
        expect_type=LookupChildType.ARTIFACT,
    ).artifact
    return locate_debian_source_package(configuration_key, artifact)


@functools.lru_cache(maxsize=100)
def source_package_data(
    workflow: "Workflow[Any, Any]",
    *,
    configuration_key: str = "source_artifact",
) -> DebianSourcePackage:
    """Return source package artifact data for the workflow."""
    return SourcePackage.create_data(
        source_package(workflow, configuration_key=configuration_key).data
    )


def lookup_result_artifact_category(result: LookupResult) -> str:
    """
    Get artifact category from result of looking up an artifact.

    The result may be either an artifact or a promise.
    """
    if (
        result.result_type == LookupResultType.ARTIFACT
        and result.artifact is not None
    ):
        return result.artifact.category
    elif (
        result.result_type == LookupResultType.BARE
        and result.collection_item is not None
        and result.collection_item.category == BareDataCategory.PROMISE
    ):
        return DebusinePromise(**result.collection_item.data).promise_category
    else:
        raise ValueError(
            f"Cannot determine artifact category for lookup result: {result}"
        )


def lookup_result_architectures(result: LookupResult) -> set[str]:
    """Get binary package architectures present within LookupResult."""
    if result.artifact is not None:
        artifact_data = result.artifact.create_data()
        if not isinstance(artifact_data, get_args(ArtifactDataWithArch)):
            raise ValueError(f"Unexpected type: {type(artifact_data).__name__}")
        return get_architectures(artifact_data)
    elif result.collection_item is not None:
        architecture = result.collection_item.data.get("architecture")
        if type(architecture) is not str:
            raise ValueError(
                f"Cannot determine architecture for lookup result: {result}"
            )
        return {architecture}
    else:
        raise ValueError(
            "Unexpected result: must have collection_item or artifact"
        )


class ArtifactHasNoBinaryPackageName(Exception):
    """Raised if it's not possible to determine the artifact's binary name."""


def lookup_result_binary_package_name(result: LookupResult) -> str:
    """Get binary package name from result of looking up an artifact."""
    binary_package_name: str | None

    if result.artifact is not None:
        artifact_data = result.artifact.create_data()
        match artifact_data:
            case DebianBinaryPackage():
                binary_package_name = artifact_data.deb_fields.get("Package")
            case _:
                raise ArtifactHasNoBinaryPackageName(f"{type(artifact_data)}")
    elif result.collection_item is not None:
        binary_package_name = result.collection_item.data.get(
            "binary_package_name"
        )
    else:
        raise ValueError(
            "Unexpected result: must have collection_item or artifact"
        )

    if not isinstance(binary_package_name, str):
        raise ValueError(
            f"Cannot determine binary package name for lookup result: {result}"
        )

    return binary_package_name


def filter_artifact_lookup_by_arch(
    workflow: "Workflow[Any, Any]",
    lookup: LookupMultiple,
    architectures: set[str],
) -> LookupMultiple:
    """Filter an artifact lookup by architecture."""
    results = workflow.work_request.lookup_multiple(
        lookup, expect_type=LookupChildType.ARTIFACT_OR_PROMISE
    )
    relevant: list[LookupSingle] = []
    for result in results:
        arch_in_lookup = lookup_result_architectures(result)
        if arch_in_lookup.issubset(architectures):
            relevant.append(
                result.reconstruct(
                    workflow_root=workflow.work_request.workflow_root
                )
            )
        elif arch_in_lookup.intersection(architectures):
            # Uploads are the only artifacts with multiple architectures at
            # this time.
            assert result.artifact is not None
            assert result.artifact.category == ArtifactCategory.UPLOAD
            for binary_package in follow_artifact_relations(
                result.artifact,
                ArtifactRelation.Relations.EXTENDS,
                ArtifactCategory.BINARY_PACKAGE,
            ):
                pkg_data = cast(
                    DebianBinaryPackage, binary_package.create_data()
                )
                pkg_architectures = get_architectures(pkg_data)
                if pkg_architectures.issubset(architectures):
                    relevant.append(f"{binary_package.id}@artifacts")

    return LookupMultiple.model_validate(sorted(relevant, key=str))


def get_lookup_architectures(
    workflow: "Workflow[Any, Any]", lookup: LookupMultiple
) -> set[str]:
    """
    Return set with all the architectures in the artifacts from the lookup.

    The architectures are extracted from each lookup result using
    :py:func:`lookup_result_architectures`.
    """
    results = workflow.work_request.lookup_multiple(
        lookup, expect_type=LookupChildType.ARTIFACT_OR_PROMISE
    )

    architectures: set[str] = set()
    for result in results:
        architectures |= lookup_result_architectures(result)
    return architectures


def follow_artifact_relations(
    artifact: Artifact,
    relation_type: ArtifactRelation.Relations,
    category: ArtifactCategory,
) -> QuerySet[Artifact]:
    """Follow relations from artifact to find artifacts of category."""
    return Artifact.objects.filter(
        targeted_by__artifact=artifact,
        targeted_by__type=relation_type,
        category=category,
    ).order_by("id")


def locate_debian_source_package(
    configuration_key: str, artifact: Artifact
) -> Artifact:
    """
    Find a :artifact:`debian:source-package` artifact for a workflow.

    :param configuration_key: The key in the workflow's task data from which
      the given artifact was looked up.
    :param artifact: A :artifact:`debian:source-package` or
      :artifact:`debian:upload` artifact.
    :return: If ``artifact`` is a
      :artifact:`debian:source-package`, return it; if it is a
      :artifact:`debian:upload`, return the related
      :artifact:`debian:source-package`.
    """
    ensure_artifact_categories(
        configuration_key=configuration_key,
        category=artifact.category,
        expected=[ArtifactCategory.SOURCE_PACKAGE, ArtifactCategory.UPLOAD],
    )
    match artifact.category:
        case ArtifactCategory.SOURCE_PACKAGE:
            return artifact
        case ArtifactCategory.UPLOAD:
            source_packages = follow_artifact_relations(
                artifact,
                ArtifactRelation.Relations.EXTENDS,
                ArtifactCategory.SOURCE_PACKAGE,
            )
            try:
                return source_packages.get()
            except Artifact.DoesNotExist:
                raise TaskConfigError(
                    f"Unable to find an artifact of category "
                    f"{ArtifactCategory.SOURCE_PACKAGE} with a relationship "
                    f"of type {ArtifactRelation.Relations.EXTENDS} from "
                    f'"{artifact}"'
                )
            except Artifact.MultipleObjectsReturned:
                raise TaskConfigError(
                    f"Multiple artifacts of category "
                    f"{ArtifactCategory.SOURCE_PACKAGE} with a relationship "
                    f"of type {ArtifactRelation.Relations.EXTENDS} from "
                    f'"{artifact}" found'
                )
        case _ as unreachable:  # pragma: no cover
            raise AssertionError(f"Unexpected artifact category: {unreachable}")


def locate_debian_source_package_lookup(
    workflow: "Workflow[Any, Any]", configuration_key: str, lookup: LookupSingle
) -> LookupSingle:
    """
    Return a lookup to a :artifact:`debian:source-package`.

    If the specified lookup returns a :artifact:`debian:source-package`,
    return it.  If it returns a :artifact:`debian:upload`, find the related
    :artifact:`debian:source-package` and return a lookup to it.
    """
    artifact = workflow.work_request.lookup_single(
        lookup, expect_type=LookupChildType.ARTIFACT
    ).artifact
    if artifact.category == ArtifactCategory.UPLOAD:
        source_package = locate_debian_source_package(
            configuration_key, artifact
        )
        return f"{source_package.id}@artifacts"
    ensure_artifact_categories(
        configuration_key=configuration_key,
        category=artifact.category,
        expected=[ArtifactCategory.SOURCE_PACKAGE],
    )
    return lookup


def locate_debian_binary_packages(
    configuration_key: str, artifacts: Sequence[Artifact]
) -> list[Artifact]:
    """
    Find :artifact:`debian:binary-package` artifacts for a workflow.

    :param configuration_key: The key in the workflow's task data from which
      the given artifacts were looked up.
    :param artifacts: A sequence of :artifact:`debian:binary-package` or
      :artifact:`debian:upload` artifacts.
    :return: A list of artifacts: for each element of ``artifacts``, if it
      is a :artifact:`debian:binary-package`, return it; if it is a
      :artifact:`debian:upload`, return all the related
      :artifact:`debian:binary-package` artifacts.
    """
    binary_packages: list[Artifact] = []
    for i, artifact in enumerate(artifacts):
        ensure_artifact_categories(
            configuration_key=f"{configuration_key}[{i}]",
            category=artifact.category,
            expected=[ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD],
        )
        match artifact.category:
            case ArtifactCategory.BINARY_PACKAGE:
                binary_packages.append(artifact)
            case ArtifactCategory.UPLOAD:
                binary_packages += follow_artifact_relations(
                    artifact,
                    ArtifactRelation.Relations.EXTENDS,
                    ArtifactCategory.BINARY_PACKAGE,
                )
            case _ as unreachable:  # pragma: no cover
                raise AssertionError(
                    f"Unexpected artifact category: {unreachable}"
                )
    return binary_packages


def get_source_package_names(
    results: Sequence[LookupResult],
    *,
    configuration_key: str,
    artifact_expected_categories: AbcCollection[ArtifactCategory],
) -> list[str]:
    """
    Return a sorted list of source package names from results.

    It ensures that:

    - The :py:class:`LookupResult` objects contain either an artifact or
      promise.
    - Artifacts belong to the artifact_expected_categories.
    - If :py:class:`LookupResult` is a promise: extracts the name from the
      promise data ``source_package_name``.

    :param results: A sequence of :py:class:`LookupResult` objects
      representing artifacts to be processed. Each entry is expected to be
      either an artifact or a promise.
    :param configuration_key: A string used by
      :py:meth:`ensure_artifact_categories` for the exception
      message.
    :param artifact_expected_categories: valid :py:class:`ArtifactCategory`
      that artifacts must belong to.
    :return: A sorted list of source package names.
    """
    source_package_names = set()

    for result in results:
        # lookup_multiple expect_type: only artifacts or promises
        match result.result_type:
            case LookupResultType.ARTIFACT:
                assert result.artifact is not None
                category = result.artifact.category

                ensure_artifact_categories(
                    configuration_key=configuration_key,
                    category=category,
                    expected=artifact_expected_categories,
                )
                artifact_data = result.artifact.create_data()

                assert isinstance(
                    artifact_data,
                    (
                        DebianSourcePackage,
                        DebianUpload,
                        DebianBinaryPackage,
                        DebianPackageBuildLog,
                    ),
                )
                source_package_names.add(get_source_package_name(artifact_data))

            case _:
                # Makes coverage happy
                # It's a promise.
                assert result.result_type == LookupResultType.BARE
                assert result.collection_item is not None

                ensure_artifact_categories(
                    configuration_key=configuration_key,
                    category=result.collection_item.data["promise_category"],
                    expected=artifact_expected_categories,
                )

                if (
                    package_name := result.collection_item.data.get(
                        "source_package_name"
                    )
                ) is not None:
                    source_package_names.add(package_name)

    return sorted(source_package_names)


def get_available_architectures(
    workflow: "Workflow[Any, Any]", *, vendor: str, codename: str
) -> set[str]:
    """Get architectures available for use with this vendor/codename."""
    architectures = set()

    for result in workflow.work_request.lookup_multiple(
        LookupMultiple.model_validate(
            {"collection": vendor, "data__codename": codename}
        ),
        default_category=CollectionCategory.ENVIRONMENTS,
        expect_type=LookupChildType.ARTIFACT,
    ):
        architectures.add(result.artifact.data.get("architecture"))

    if not architectures:
        raise TaskConfigError(
            f"Unable to find any environments for {vendor}:{codename}"
        )

    architectures.add("all")

    return architectures


def configure_for_overlay_suite(
    workflow: "Workflow[Any, Any]",
    *,
    extra_repositories: Sequence[ExtraRepository] | None,
    vendor: str,
    codename: str,
    environment: LookupSingle,
    backend: BackendType,
    architecture: str,
    try_variant: str,
) -> list[ExtraRepository] | None:
    """Return any needed extra repository to use an overlay suite."""
    match (vendor, codename):
        case ("debian", "experimental"):
            pass
        case _:
            return (
                None if extra_repositories is None else list(extra_repositories)
            )

    if extra_repositories is None:
        extra_repositories = []

    environment_artifact = workflow.work_request.lookup_environment(
        environment,
        architecture=architecture,
        backend=backend,
        default_category=CollectionCategory.ENVIRONMENTS,
        try_variant=try_variant,
    ).artifact.create_data()
    assert isinstance(environment_artifact, DebianSystemTarball)
    mirror = environment_artifact.mirror
    components = environment_artifact.components

    return [
        *extra_repositories,
        ExtraExternalRepository(
            url=mirror, suite=codename, components=components
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UEFI Secure Boot signing utility for the Debusine signing service."""

import os
import subprocess
from pathlib import Path
from typing import BinaryIO, TYPE_CHECKING

from debusine.signing.models import ProtectedKey

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer

    ReadableBuffer  # fake usage for vulture


def sbsign(
    private_key: ProtectedKey,
    public_key: "ReadableBuffer",
    data_path: Path,
    signature_path: Path,
    detached: bool = False,
    log_file: BinaryIO | None = None,
) -> None:
    """Sign a UEFI boot image using `sbsign`."""
    with private_key.available(public_key) as available:
        cmd: list[str | os.PathLike[str]] = ["sbsign", *available.sbsign_args]
        if detached:
            cmd.append("--detached")
        cmd.extend(["--output", signature_path, data_path])
        subprocess.run(cmd, check=True, stdout=log_file, stderr=log_file)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""API for provisioning and managing AWS EC2 cloud workers."""

import logging
from functools import lru_cache
from typing import Any, Literal, TYPE_CHECKING, cast, override

import boto3
from botocore.exceptions import ClientError
from django.utils import timezone

from debusine.assets import AWSProviderAccountData, CloudProvidersType
from debusine.db.models.auth import Token
from debusine.server.worker_pools.base import WorkerPoolInterface
from debusine.server.worker_pools.cloud_init import (
    CloudInitConfig,
    FilesystemSetup,
    worker_bootstrap_cloud_init,
)
from debusine.server.worker_pools.exceptions import InstanceLaunchException
from debusine.server.worker_pools.models import (
    AWSEC2InstanceMarketType,
    AWSEC2LaunchTemplate,
    AWSEC2WorkerPoolSpecification,
)

if TYPE_CHECKING:
    from mypy_boto3_ec2.client import EC2Client
    from mypy_boto3_ec2.type_defs import (
        FleetLaunchTemplateConfigRequestTypeDef,
        InstanceRequirementsRequestTypeDef,
        LaunchTemplateBlockDeviceMappingRequestTypeDef,
        LaunchTemplateInstanceNetworkInterfaceSpecificationRequestTypeDef,
        LaunchTemplateTagSpecificationRequestTypeDef,
        RequestLaunchTemplateDataTypeDef,
        SpotOptionsRequestTypeDef,
    )

    from debusine.db.models import Worker

    # fake usage for vulture
    EC2Client
else:
    # They're all TypedDicts
    FleetLaunchTemplateConfigRequestTypeDef = dict
    InstanceRequirementsRequestTypeDef = dict
    LaunchTemplateBlockDeviceMappingRequestTypeDef = dict
    LaunchTemplateTagSpecificationRequestTypeDef = dict
    RequestLaunchTemplateDataTypeDef = dict
    SpotOptionsRequestTypeDef = dict
    LaunchTemplateInstanceNetworkInterfaceSpecificationRequestTypeDef = dict

log = logging.getLogger(__name__)


class AWSEC2WorkerPool(
    WorkerPoolInterface, cloud_provider=CloudProvidersType.AWS
):
    """Amazon EC2 Worker Pool."""

    specifications: AWSEC2WorkerPoolSpecification
    client: "EC2Client"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize an AWSEC2WorkerPool."""
        super().__init__(*args, **kwargs)
        provider_account_data = self.worker_pool.provider_account.data_model
        assert isinstance(provider_account_data, AWSProviderAccountData)
        self.client = self._make_client(provider_account_data)

    @override
    def launch_worker(self, worker: "Worker") -> None:
        assert worker.activation_token

        self._terminate_existing_instances(worker.worker_pool_data)

        specs = self.specifications
        launch_template_ids = []
        try:
            # We support multiple templates, to allow finding the cheapest
            # instance even if different instances require different
            # configuration
            for id_, launch_template in enumerate(specs.launch_templates):
                name = f"debusine_worker_{worker.name}_{id_}"
                try:
                    # If there was a left-over template from before, delete it
                    self.client.delete_launch_template(LaunchTemplateName=name)
                except ClientError:
                    pass
                launch_template_ids.append(
                    self._create_launch_template(
                        name,
                        self._launch_template_data(
                            launch_template,
                            worker.name,
                            worker.activation_token,
                        ),
                    )
                )
            instance_id = self._create_fleet(
                launch_template_ids,
                instance_market_type=specs.instance_market_type,
                worker_name=worker.name,
                max_spot_price_per_hour=specs.max_spot_price_per_hour,
            )
        finally:
            for template_id in launch_template_ids:
                self.client.delete_launch_template(LaunchTemplateId=template_id)

        worker.instance_created_at = timezone.now()
        worker.worker_pool_data = {"instance-id": instance_id}
        worker.save()

    @override
    def terminate_missing(self) -> None:
        instance_ids: dict[str, int] = {
            instance_id: worker_id
            for (
                worker_id,
                instance_id,
            ) in self.worker_pool.worker_set.active().values_list(
                "id", "worker_pool_data__instance-id"
            )
        }
        if not instance_ids:
            return

        paginator = self.client.get_paginator("describe_instances")
        for page in paginator.paginate(InstanceIds=list(instance_ids.keys())):
            for reservation in page["Reservations"]:
                for instance in reservation["Instances"]:
                    if instance["State"]["Name"] in ("pending", "running"):
                        instance_ids.pop(instance["InstanceId"], None)

        for instance_id, worker_id in instance_ids.items():
            worker = self.worker_pool.worker_set.get(id=worker_id)
            log.warning(
                "Worker %s (instance %s) has gone missing. Terminating.",
                worker.name,
                instance_id,
            )
            self.worker_pool.terminate_worker(worker)

    def _make_client(
        self, provider_account_data: AWSProviderAccountData
    ) -> "EC2Client":
        """Make an EC2 client for this provider account."""
        return boto3.client(
            "ec2",
            region_name=provider_account_data.configuration.region_name,
            endpoint_url=provider_account_data.configuration.ec2_endpoint_url,
            aws_access_key_id=provider_account_data.credentials.access_key_id,
            aws_secret_access_key=(
                provider_account_data.credentials.secret_access_key
            ),
        )

    @override
    def _terminate_existing_instances(
        self, worker_pool_data: dict[str, Any] | None
    ) -> None:
        """
        If worker_pool_data references an instance-id, terminate it.

        We are often working with spot instances that vanish unexpectedly, just
        ignore failures for unknown instances.
        """
        if worker_pool_data is None:
            return
        if existing_instance := worker_pool_data.get("instance-id", None):
            try:
                self.client.terminate_instances(InstanceIds=[existing_instance])
            except ClientError as e:
                if e.response["Error"]["Code"] == "InvalidInstanceID.NotFound":
                    return
                raise

    def _cloud_init_config(
        self,
        launch_template: AWSEC2LaunchTemplate,
        worker_name: str,
        activation_token: Token,
    ) -> CloudInitConfig:
        """Generate cloud-init configuration for a new instance."""
        config = worker_bootstrap_cloud_init(
            worker_name,
            activation_token,
            debusine_install_source=self.specifications.debusine_install_source,
            debusine_install_source_custom=self.specifications.debusine_install_source_custom,
            debian_release=self.specifications.debian_release,
            apt_authentication=self.get_apt_authentication(),
        )
        if launch_template.swap_size:
            config.device_aliases["swap"] = "/dev/xvdb"
            config.fs_setup.append(
                FilesystemSetup(device="swap", filesystem="swap")
            )
            config.mounts.append(
                [
                    "swap",
                    "none",
                    "swap",
                    "sw,nofail,x-systemd.requires=cloud-init.service",
                ]
            )
        return config

    @lru_cache
    def _lookup_image_root_device_name(self, image_id: str) -> str:
        """Look up the DeviceName of the root filesystem in the image."""
        response = self.client.describe_images(ImageIds=[image_id])
        images = response["Images"]
        if len(images) != 1:
            raise ValueError(
                f"Found {len(images)} images instead of the expected 1"
            )
        image = images[0]

        if image["RootDeviceType"] != "ebs":
            raise ValueError(f"Image {image_id} is not EBS-backed")
        return image["RootDeviceName"]

    def _launch_template_data(
        self,
        launch_template: AWSEC2LaunchTemplate,
        worker_name: str,
        activation_token: Token,
    ) -> RequestLaunchTemplateDataTypeDef:
        """
        Generate a low-level LaunchTemplateData from launch_template.

        AWSEC2LaunchTemplate includes some low-level properties directly, but
        also provides higher level abstractions.
        """
        block_device_mappings: list[
            LaunchTemplateBlockDeviceMappingRequestTypeDef
        ] = []
        if launch_template.root_device_size:
            block_device_mappings.append(
                {
                    "DeviceName": self._lookup_image_root_device_name(
                        launch_template.ImageId
                    ),
                    "Ebs": {
                        "VolumeSize": launch_template.root_device_size,
                        "DeleteOnTermination": True,
                        "VolumeType": "gp3",
                    },
                }
            )
        if launch_template.swap_size:
            block_device_mappings.append(
                {
                    "DeviceName": "/dev/xvdb",
                    "Ebs": {
                        "VolumeSize": launch_template.swap_size,
                        "DeleteOnTermination": True,
                        "VolumeType": "gp3",
                    },
                },
            )
        cloud_init_config = self._cloud_init_config(
            launch_template, worker_name, activation_token
        )

        tag_specifications: list[
            LaunchTemplateTagSpecificationRequestTypeDef
        ] = []
        if launch_template.tags:
            tag_specifications.append(
                {
                    "ResourceType": "instance",
                    "Tags": [
                        {"Key": key, "Value": value}
                        for key, value in launch_template.tags.items()
                    ],
                }
            )

        data: RequestLaunchTemplateDataTypeDef = {
            "EbsOptimized": launch_template.EbsOptimized,
            "BlockDeviceMappings": block_device_mappings,
            "ImageId": launch_template.ImageId,
            "MetadataOptions": {
                "HttpTokens": "required",
            },
            "UserData": cloud_init_config.as_base64_str(),
            "TagSpecifications": tag_specifications,
        }

        if launch_template.InstanceRequirements:
            data["InstanceRequirements"] = cast(
                InstanceRequirementsRequestTypeDef,
                launch_template.InstanceRequirements.model_dump(
                    exclude_unset=True
                ),
            )
        else:
            # mypy_boto3 types this as a very long set of Literal options
            data["InstanceType"] = launch_template.InstanceType  # type: ignore[typeddict-item]

        if launch_template.KeyName:
            data["KeyName"] = launch_template.KeyName
        if launch_template.NetworkInterfaces:
            data["NetworkInterfaces"] = [
                cast(
                    LaunchTemplateInstanceNetworkInterfaceSpecificationRequestTypeDef,  # noqa: E501
                    interface.model_dump(exclude_none=True),
                )
                for interface in launch_template.NetworkInterfaces
            ]

        return data

    def _create_launch_template(
        self,
        name: str,
        launch_template_data: RequestLaunchTemplateDataTypeDef,
    ) -> str:
        """
        Create an EC2 Launch template, and return its ID.

        We have to create server-side named templates, to be able to pass
        UserData: https://github.com/aws/aws-sdk/issues/528
        """
        response = self.client.create_launch_template(
            LaunchTemplateName=name,
            LaunchTemplateData=launch_template_data,
        )
        return response["LaunchTemplate"]["LaunchTemplateId"]

    def _create_fleet(
        self,
        launch_template_ids: list[str],
        instance_market_type: AWSEC2InstanceMarketType,
        worker_name: str,
        max_spot_price_per_hour: float | None,
    ) -> str:
        """Create an EC2 instance using create_fleet, and return its ID."""
        if not launch_template_ids:
            raise ValueError("No launch_template_ids were provided")
        launch_template_configs: list[
            FleetLaunchTemplateConfigRequestTypeDef
        ] = [
            {
                "LaunchTemplateSpecification": {
                    "LaunchTemplateId": launch_template_id,
                    "Version": "$Latest",
                },
            }
            for launch_template_id in launch_template_ids
        ]
        spot_options: SpotOptionsRequestTypeDef = {}
        if (
            instance_market_type == AWSEC2InstanceMarketType.SPOT
            and max_spot_price_per_hour
        ):
            spot_options["MaxTotalPrice"] = str(max_spot_price_per_hour)

        response = self.client.create_fleet(
            SpotOptions=spot_options,
            LaunchTemplateConfigs=launch_template_configs,
            TargetCapacitySpecification={
                "TotalTargetCapacity": 1,
                "DefaultTargetCapacityType": cast(
                    Literal["on-demand", "spot"], str(instance_market_type)
                ),
            },
            Type="instant",
            TagSpecifications=[
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "Name", "Value": worker_name}],
                },
            ],
        )
        # No exception is raised by boto3 on an error, it's a 200
        for error in response.get("Errors", []):
            if error["ErrorCode"] == "InsufficientInstanceCapacity":
                # These seem to be non-fatal
                log.warning(
                    "Amazon EC2 reported %s, when attempting to launch %s: %s",
                    error["ErrorCode"],
                    worker_name,
                    error["ErrorMessage"],
                )
                continue
            raise InstanceLaunchException(
                f"Failure to launch fleet: {error['ErrorCode']}: "
                f"{error['ErrorMessage']}"
            )
        instance_ids = []
        for instance in response["Instances"]:
            instance_ids += instance["InstanceIds"]
        if len(instance_ids) == 0:
            raise InstanceLaunchException("No instances launched")
        assert len(instance_ids) == 1
        return instance_ids[0]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine dput-ng hooks."""

from typing import Any
from urllib.parse import quote, urljoin

from dput.changes import Changes
from dput.core import logger
from dput.exceptions import HookException
from dput.interface import AbstractInterface

from debusine.client.dput_ng.dput_ng_utils import make_debusine_client
from debusine.client.models import CreateWorkflowRequest


class NoWorkflowError(HookException):  # type: ignore[misc]
    """No debusine workflow is configured for this distribution."""


def choose_workflow(changes: Changes, profile: dict[str, Any]) -> str:
    """
    Select the debusine workflow to create after the upload.

    :raises NoWorkflowError: if we do not know what workflow to create.
    """
    if isinstance((workflow_name := profile.get("debusine_workflow")), str):
        return workflow_name
    distribution = changes["Distribution"]
    try:
        key = distribution
        if key not in profile["debusine_workflows_by_distribution"]:
            key = "*"
        workflow_name = profile["debusine_workflows_by_distribution"][key]
        assert isinstance(workflow_name, str)
        workflow_name = workflow_name.format(distribution=distribution)
        return workflow_name
    except KeyError:
        raise NoWorkflowError(
            f"No debusine workflow is configured for {distribution!r}"
        )


def check_workflow(
    changes: Changes,
    profile: dict[str, Any],
    interface: AbstractInterface,  # noqa: ARG001, U100
) -> None:
    """Check that we know what debusine workflow to create."""
    choose_workflow(changes, profile)


def create_workflow(
    changes: Changes,
    profile: dict[str, Any],
    interface: AbstractInterface,  # noqa: ARG001, U100
) -> None:
    """Create a workflow based on an upload to debusine."""
    scope = profile["debusine_scope"]
    workspace = profile["debusine_workspace"]
    try:
        workflow_template_name = choose_workflow(changes, profile)
    except NoWorkflowError:
        # Give up; this should already have been checked by a pre-upload hook.
        return
    task_data = profile.get("debusine_workflow_data", {})
    for key in task_data:
        if task_data[key] == "@UPLOAD@":
            task_data[key] = profile["debusine_artifact_ids"]["changes"]
        if task_data[key] == "@DISTRIBUTION@":
            task_data[key] = changes["Distribution"]
    debusine = make_debusine_client(profile)
    workflow = CreateWorkflowRequest(
        template_name=workflow_template_name,
        workspace=workspace,
        task_data=task_data,
    )
    try:
        workflow_created = debusine.workflow_create(workflow)
    except Exception as e:
        raise HookException(
            f"Failed to create workflow {workflow_template_name}: {e}"
        ) from e
    logger.info(
        "Created workflow: %s",
        # TODO: Ideally we wouldn't have to hardcode the URL structure here.
        urljoin(
            debusine.base_api_url,
            f"/{quote(scope)}/{quote(workspace)}/work-request"
            f"/{workflow_created.id}/",
        ),
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Extract data from docstrings and format it as RST."""

import inspect
import re

from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst import Directive
from sphinx.ext.autodoc.importer import import_object
from sphinx.util.typing import OptionSpec


class DynamicDataDirective(Directive):
    """
    Sphinx directive to document dynamic data.

    To use, in an .rst file:

    .. code-block:: rst

      .. dynamic_data::
         :method: debusine.tasks.sbuild::Sbuild.compute_dynamic_data

    In the Sbuild.compute_dynamic_data method include in its docstring:

    .. code-block:: rst

        Computes the dynamic data for the task resolving artifacts.

        :subject: source package that is building
        :runtime_context: the context
    """

    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False

    option_spec: OptionSpec = {
        "method": str,
    }

    def run(self) -> list[Node]:
        """Run the dynamic data directive."""
        method_path = self.options.get("method")
        if not method_path or "::" not in method_path:
            return [
                self.state_machine.reporter.error(
                    '"dynamic_data" directive requires "method" option '
                    'in the format of "method: module_path::Class.method"',
                    line=self.lineno,
                )
            ]

        module_name, obj_path = method_path.split("::", 1)

        try:
            _, _, _, method = import_object(
                module_name, obj_path.split("."), objtype="method"
            )
        except ModuleNotFoundError:
            return [
                self.state_machine.reporter.error(
                    f'Cannot import "{method_path}". '
                    f'Expected "module::Class.method"',
                    line=self.lineno,
                )
            ]

        docstring = inspect.getdoc(method)

        if docstring is None:
            return [
                self.state_machine.reporter.error(
                    f"No docstring found in '{method_path}'",
                    line=self.lineno,
                )
            ]

        parsed = self._extract_dynamic_data_fields(docstring)

        if len(parsed) == 0:
            return [
                self.state_machine.reporter.error(
                    'No "Dynamic Data:" block or field keys found in '
                    f'"{method_path}"',
                    line=self.lineno,
                )
            ]

        return self._generate_dynamic_data_rst(parsed)

    @staticmethod
    def _extract_dynamic_data_fields(docstring: str) -> dict[str, str]:
        """
        Extract dynamic data fields from a docstring using field lists.

        :return: dictionary with dynamic data field names and descriptions.
        """
        dynamic_data = {}
        lines = docstring.splitlines()

        field_indentation_level = 0
        current_key: str | None = None

        for line in lines:
            stripped = line.strip()

            current_indentation_level = len(line) - len(line.lstrip(" "))

            if (
                current_key is not None
                and current_indentation_level > field_indentation_level
            ):
                dynamic_data[current_key] += f" {stripped}"
                continue

            match = re.match(r":(\w+):\s*(.*)", stripped)
            if match:
                key, value = match.groups()
                if key in ["returns", "return"]:
                    # Skip "returns" / "return" which is likely to be what the
                    # method return, not a configuration key
                    continue

                current_key = key
                dynamic_data[current_key] = value
                field_indentation_level = current_indentation_level

        return dynamic_data

    def _generate_dynamic_data_rst(
        self, dynamic_data: dict[str, str]
    ) -> list[nodes.bullet_list]:
        """Generate a Sphinx field list from extracted dynamic data."""
        bullet_list = nodes.bullet_list()

        for key, value in dynamic_data.items():
            contents = nodes.paragraph()

            contents += nodes.strong(text=f"{key}: ")

            inline_nodes, _ = self.state.inline_text(value, self.lineno)
            contents += inline_nodes

            list_item = nodes.list_item()
            list_item += contents

            bullet_list.append(list_item)

        return [bullet_list]
